# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Bridge for using cclib data in horton (http://theochem.github.io/horton)."""

import numpy
from cclib.parser.data import ccData
from cclib.parser.utils import convertor, find_package

_found_iodata = find_package("iodata")

# Detect whether iodata (part of horton 3) is present or not
# Horton 3 is divided into smaller (sub)packages that each take different functionalities.
if _found_iodata:
    from iodata import IOData
    from iodata.orbitals import MolecularOrbitals


def check_horton():
    if not _found_iodata:
        raise ImportError("You must install `horton` to use this function.")


def makehorton(data):
    """ Create horton IOData object from ccData object """

    check_horton()
    attributes = {}

    # In horton 3 IOData, inputs are type-checked -- thus the bridge also verifies the types.
    # if coordinates are known; numpy.ndarray (size natom-by-3)
    if hasattr(data, "atomcoords"):
        attributes["atcoords"] = convertor(
            numpy.asanyarray(data.atomcoords)[-1], "Angstrom", "bohr"
        )
    # if atomic numbers are known; numpy.ndarray (size natom)
    if hasattr(data, "atomnos"):
        attributes["atnums"] = numpy.asanyarray(data.atomnos)
    # if orbital coefficients known; iodata.orbitals.MolecularOrbitals
    if hasattr(data, "mocoeffs"):
        # First build a dictionary of inputs that will be passed to the constructor of
        # horton3's MolecularOrbitals object.
        moattr = {
            "kind": "restricted",
            "norba": data.nmo,
            "norbb": None,
            # In horton 3, occupation in MOs are represented as 1's (occupied) and
            # 0's (unoccupied). Beta orbitals follow directly after alpha orbitals, forming
            # 1D array.
            "occs": numpy.concatenate(
                (numpy.ones(data.homos[0]), numpy.zeros(data.nmo - data.homos[0])),
            ),
            "coeffs": data.mocoeffs[0].T,
            "energies": None,
            "irreps": None,
        }
        # and if unrestricted:
        if len(data.mocoeffs) == 2:
            moattr["kind"] = "unrestricted"
            moattr["norbb"] = data.nmo
            moattr["coeffs"] = numpy.concatenate(
                (moattr["coeffs"], data.mocoeffs[1].T),
                axis=1
            )
            moattr["occs"] = numpy.concatenate(
                    (
                        moattr["occs"],
                        numpy.ones(data.homos[1]),
                        numpy.zeros(data.nmo - data.homos[1]),
                    )
            )
        # Then construct MolecularOrbitals object
        attributes["mo"] = MolecularOrbitals(**moattr)
    # if multiplicity known; float / should not be set when mocoeffs present
    # Refer to IOData code:
    # https://github.com/theochem/iodata/blob/b36513d162f99b57264005583701c6987037839c/iodata/iodata.py#L174
    if (
        hasattr(data, "mult")
        and not hasattr(data, "mocoeffs")
    ):
        attributes["spinpol"] = data.mult - 1  # horton has 2S+1, iodata has 2S
    # if pseudopotentials exist; numpy.ndarray (size natom)
    if hasattr(data, "coreelectrons"):
        attributes["atcorenums"] = data.atomnos - numpy.asanyarray(data.coreelectrons)
    # if mulliken charges are known; dict of numpy.ndarrays (size natom)
    if hasattr(data, "atomcharges"):
        attributes["atcharges"] = data.atomcharges

    return IOData(**attributes)  # Pass collected attributes into IOData constructor


def makecclib(iodat):
    """ Create cclib ccData object from horton IOData object """

    check_horton()
    attributes = {}

    # Horton 3 IOData class uses attr and does not have __dict__.
    # In horton 3, some attributes have a default value of None.
    # Therefore, second hasattr statement is needed for mo attribute.
    if hasattr(iodat, "atcoords"):
        # cclib parses the whole history of coordinates in the list, horton keeps the last one.
        attributes["atomcoords"] = [convertor(iodat.atcoords, "bohr", "Angstrom")]
    if hasattr(iodat, "mo") and hasattr(iodat.mo, "norba"):
        # MO coefficient should be transposed to match the dimensions.
        attributes["mocoeffs"] = [iodat.mo.coeffs[:, :iodat.mo.norba].T]
        if iodat.mo.kind == "unrestricted":
            attributes["mocoeffs"].append(iodat.mo.coeffs[:, iodat.mo.norba:].T)
    if hasattr(iodat, "spinpol") and isinstance(iodat.spinpol, int):
        # IOData stores 2S, ccData stores 2S+1.
        attributes["mult"] = iodat.spinpol + 1
    if hasattr(iodat, "atnums"):
        attributes["atnums"] = numpy.asanyarray(iodat.atnums)
    if hasattr(iodat, "atcorenums") and isinstance(iodat.atnums, numpy.ndarray):
        # cclib stores num of electrons screened out by pseudopotential
        # horton stores num of electrons after applying pseudopotential
        attributes["coreelectrons"] = numpy.asanyarray(iodat.atnums) - numpy.asanyarray(
            iodat.atcorenums
        )
    if hasattr(iodat, "atcharges"):
        attributes["atomcharges"] = iodat.atcharges

    return ccData(attributes)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Unit tests for wfx writer."""

import os
import unittest
import numpy as np

import cclib

from cclib.io.filewriter import MissingAttributeError
from cclib.io.wfxwriter import _section
from cclib.io.wfxwriter import _list_format
import pytest

__filedir__ = os.path.dirname(__file__)
__filepath__ = os.path.realpath(__filedir__)
__datadir__ = os.path.join(__filepath__, "..", "..")


class WFXTest(unittest.TestCase):

    def test_missing_attribute_error(self):
        """Check if MissingAttributeError is raised as expected."""
        fpath = os.path.join(__datadir__,
                             "data/GAMESS/basicGAMESS-US2017/C_bigbasis.out")
        required_attrs = ('atomcoords', 'atomnos', 'gbasis', 'charge',
                          'homos', 'mult', 'mocoeffs')
        for attr in required_attrs:
            data = cclib.io.ccread(fpath)
            delattr(data, attr)

            # WFX files cannot be written if required attrs are missing.
            with pytest.raises(MissingAttributeError):
                cclib.io.wfxwriter.WFXWriter(data)

    def test_no_of_prims(self):
        """Check if number of primitives are calculated correctly."""
        num_orb = {'s':1, 'p':3, 'd':6, 'f':10, 'g':15, 'h':21}
        gamessdir = os.path.join(__datadir__,
                                 "data/GAMESS/basicGAMESS-US2017")
        filenames = ["C_bigbasis.out", "dvb_un_sp.out"]
        filepaths = [os.path.join(gamessdir, fn) for fn in filenames]

        for fpath in filepaths:
            data = cclib.io.ccread(fpath)
            wfx = cclib.io.wfxwriter.WFXWriter(data)

            no_prims_writer = wfx._no_of_prims()
            no_prims_ccdata = 0
            for atom in data.gbasis:
                for prims in atom:
                    no_prims_ccdata += num_orb[prims[0].lower()]\
                                        * len(prims[1])

            assert no_prims_writer == no_prims_ccdata

    def test_mo_normalization(self):
        """Check if MO section is printed correctly."""
        fpath = os.path.join(__datadir__,
                             "data/GAMESS/basicGAMESS-US2017/C_bigbasis.out")
        data = cclib.io.ccread(fpath)
        wfx = cclib.io.wfxwriter.WFXWriter(data)

        normalized_mocoeffs = wfx._normalized_mocoeffs()
        if len(data.homos) > 1:
            assert len(normalized_mocoeffs) == wfx._no_electrons()
        else:
            assert len(normalized_mocoeffs) == wfx._no_of_mos()
        assert len(normalized_mocoeffs[0]) == wfx._no_of_prims()

    def test_mo_normalization_dat(self):
        """Check if MOs are normalized as expected."""
        fpath = os.path.join(__datadir__,
                             "data/GAMESS/basicGAMESS-US2017/dvb_sp.out")
        data = cclib.io.ccread(fpath)
        wfx = cclib.io.wfxwriter.WFXWriter(data)

        normalized_mocoeffs_wfx = wfx._normalized_mocoeffs()
        datfile = os.path.join(__datadir__,
                               "data/GAMESS/basicGAMESS-US2017/dvb_sp.dat")

        with open(datfile) as file:
            content = iter(file.readlines())

        normalized_mocoeffs_dat = []

        line = next(content)
        while 'TOP OF INPUT FILE FOR BADER' not in line:
            line = next(content)

        while 'END OF INPUT FILE FOR BADER' not in line:
            if 'MO' in line and 'OCC NO' in line and 'ORB. ENERGY' in line:
                line = next(content)
                mo = []
                while 'MO' not in line and 'END DATA' not in line:
                    mo.extend([float(x) for x in line.rsplit()])
                    line = next(content)
                normalized_mocoeffs_dat.append(mo)
            else:
                line = next(content)

        for mo1, mo2 in zip(normalized_mocoeffs_dat, normalized_mocoeffs_wfx):
            for coeff1, coeff2 in zip(mo1, mo2):
                assert -0.09 <= abs(coeff1) - abs(coeff2) <= 0.09


    def test_programs(self):
        """Check other programs against reference data."""
        ref_file = "data/GAMESS/basicGAMESS-US2017/dvb_sp.out"
        programs = {
            'ORCA': "data/ORCA/basicORCA4.2/dvb_sp.out",
            'NWChem': "data/NWChem/basicNWChem6.5/dvb_sp_hf.out",
            'Psi4': "data/Psi4/basicPsi4-1.3.1/dvb_sp_rhf.out",
            'GAMESS-UK': "data/GAMESS-UK/basicGAMESS-UK8.0/dvb_sp_hf.out",
            'Firefly': "data/GAMESS/basicFirefly8.0/dvb_sp.out",
        }
        fpath_ref = os.path.join(__datadir__, ref_file)
        data_ref = cclib.io.ccread(fpath_ref)
        wfx_ref = cclib.io.wfxwriter.WFXWriter(data_ref)
        norm_mat_ref = wfx_ref._norm_mat()[0]
        mos_ref = wfx_ref._no_of_mos()

        for name in programs:
            fpath_prog = os.path.join(__datadir__, programs[name])
            data_prog = cclib.io.ccread(fpath_prog)
            wfx_prog = cclib.io.wfxwriter.WFXWriter(data_prog)

            norm_mat_prog = wfx_prog._norm_mat()[0]
            mos_prog = wfx_prog._no_of_mos()

            # Check if normalization matrix are matching.
            assert round(abs(max(np.array(norm_mat_prog) - np.array(norm_mat_ref))), 5) == 0
            # Check if number of orbitals to be printed are as expected.
            assert mos_prog == mos_ref

    def test_section_printing(self):
        """Check if wfx section printing works as expected."""
        float_section = _section('Test Section', 123.456)
        expected = ['<Test Section>', ' 123.456', '</Test Section>']
        assert float_section == expected

        list_section = _section('Test Section', ['1', '2'])
        expected = ['<Test Section>', '1', '2', '</Test Section>']
        assert list_section == expected

        str_section = _section('Test Section', 'Test Data')
        expected = ['<Test Section>', ' Test Data', '</Test Section>']
        assert str_section == expected

    def test_list_format(self):
        """Check if list formatting works as expected."""
        odd_list = _list_format([1, 2, 3], 2, '%8.1E')
        odd_expected = [' 1.0E+00 2.0E+00', ' 3.0E+00']
        assert odd_list == odd_expected
        even_list = _list_format([1, 2, 3, 4], 2, '%8.1E')
        even_expected = [' 1.0E+00 2.0E+00', ' 3.0E+00 4.0E+00']
        assert even_list == even_expected


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Contains all writers for standard chemical representations."""

from cclib.io.cjsonreader import CJSON as CJSONReader
from cclib.io.cjsonwriter import CJSON as CJSONWriter
from cclib.io.cmlwriter import CML
from cclib.io.moldenwriter import MOLDEN
from cclib.io.wfxwriter import WFXWriter
from cclib.io.xyzreader import XYZ as XYZReader
from cclib.io.xyzwriter import XYZ as XYZWriter

# This allows users to type:
#   from cclib.io import ccframe
#   from cclib.io import ccopen
#   from cclib.io import ccread
#   from cclib.io import ccwrite
#   from cclib.io import URL_PATTERN
from cclib.io.ccio import ccframe
from cclib.io.ccio import ccopen
from cclib.io.ccio import ccread
from cclib.io.ccio import ccwrite



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.
"""Tools for identifying, reading and writing files and streams."""

import io
import os
import pathlib
import typing
from typing import Optional
import logging
from typing import Union
import warnings

from cclib.parser import data
from cclib.parser import logfileparser
from cclib.parser.utils import find_package

from cclib.parser.adfparser import ADF
from cclib.parser.daltonparser import DALTON
from cclib.parser.fchkparser import FChk
from cclib.parser.gamessparser import GAMESS
from cclib.parser.gamessdatparser import GAMESSDAT
from cclib.parser.gamessukparser import GAMESSUK
from cclib.parser.gaussianparser import Gaussian
from cclib.parser.jaguarparser import Jaguar
from cclib.parser.molcasparser import Molcas
from cclib.parser.molproparser import Molpro
from cclib.parser.mopacparser import MOPAC
from cclib.parser.nwchemparser import NWChem
from cclib.parser.orcaparser import ORCA
from cclib.parser.psi3parser import Psi3
from cclib.parser.psi4parser import Psi4
from cclib.parser.qchemparser import QChem
from cclib.parser.turbomoleparser import Turbomole
from cclib.parser.logfilewrapper import FileWrapper

from cclib.io import cjsonreader
from cclib.io import cjsonwriter
from cclib.io import cmlwriter
from cclib.io import moldenwriter
from cclib.io import wfxwriter
from cclib.io import xyzreader
from cclib.io import xyzwriter

_has_cclib2openbabel = find_package("openbabel")
if _has_cclib2openbabel:
    from cclib.bridge import cclib2openbabel

_has_pandas = find_package("pandas")
if _has_pandas:
    import pandas as pd

# Parser choice is triggered by certain phrases occurring the logfile. Where these
# strings are unique, we can set the parser and break. In other cases, the situation
# is a little but more complicated. Here are the exceptions:
#   1. The GAMESS trigger also works for GAMESS-UK files, so we can't break
#      after finding GAMESS in case the more specific phrase is found.
#   2. Molpro log files don't have the program header, but always contain
#      the generic string 1PROGRAM, so don't break here either to be cautious.
#   3. "MOPAC" is used in some packages like GAMESS, so match MOPAC20##
#
# The triggers are defined by the tuples in the list below like so:
#   (parser, phrases, flag whether we should break)
triggers = [

    (ADF,       ["Amsterdam Density Functional"],                   True),
    (DALTON,    ["Dalton - An Electronic Structure Program"],       True),
    (FChk,      ["Number of atoms", "I"],                           True),
    (GAMESS,    ["GAMESS"],                                         False),
    (GAMESS,    ["Firefly (PC GAMESS)"],                            True),
    (GAMESS,    ["GAMESS VERSION"],                                 True),
    (GAMESSUK,  ["G A M E S S - U K"],                              True),
    (GAMESSDAT, ["$DATA"],                                          True),
    (Gaussian,  ["Gaussian, Inc."],                                 True),
    (Jaguar,    ["Jaguar"],                                         True),
    (Molcas,    ["MOLCAS"],                                         True),
    (Molpro,    ["PROGRAM SYSTEM MOLPRO"],                          True),
    (Molpro,    ["1PROGRAM"],                                       False),
    (MOPAC,     ["MOPAC20"],                                        True),
    (NWChem,    ["Northwest Computational Chemistry Package"],      True),
    (ORCA,      ["O   R   C   A"],                                  True),
    (Psi3,      ["PSI3: An Open-Source Ab Initio Electronic Structure Package"],          True),
    (Psi4,      ["Psi4: An Open-Source Ab Initio Electronic Structure Package"],          True),
    (QChem,     ["A Quantum Leap Into The Future Of Chemistry"],    True),
    (Turbomole, ["TURBOMOLE"],                                      True),

]

readerclasses = {
    'cjson': cjsonreader.CJSON,
    'json': cjsonreader.CJSON,
    'xyz': xyzreader.XYZ,
}

writerclasses = {
    'cjson': cjsonwriter.CJSON,
    'json': cjsonwriter.CJSON,
    'cml': cmlwriter.CML,
    'molden': moldenwriter.MOLDEN,
    'wfx': wfxwriter.WFXWriter,
    'xyz': xyzwriter.XYZ,
}


class UnknownOutputFormatError(Exception):
    """Raised when an unknown output format is encountered."""


def is_xyz(inputfile: FileWrapper) -> bool:
    """Is the given inputfile actually an XYZ file?

    The only way to determine this without reading the entire file is to
    inspect the file extension.
    """
    return len(inputfile.filenames) == 1 and \
        os.path.splitext(inputfile.filenames[0])[1].lower() == ".xyz"


def guess_filetype(inputfile) -> Optional[logfileparser.Logfile]:
    """Try to guess the filetype by searching for trigger strings."""
    filetype = None
    logger = logging.getLogger("cclib")
    try:
        if isinstance(inputfile, FileWrapper) and is_xyz(inputfile):
            logger.info("Found XYZ file based on file extension")
            return filetype
        for line in inputfile:
            for parser, phrases, do_break in triggers:
                if all([line.lower().find(p.lower()) >= 0 for p in phrases]):
                    filetype = parser
                    if do_break:
                        return filetype
    except Exception:
        # guess_filetype() is expected to be quiet by default...
        logger.error("Failed to determine log file type", exc_info = True)
    
    return filetype

def sort_turbomole_outputs(fileinputs):
    """
    Sorts a list of inputs (or list of log files) according to the order
    required by the Turbomole parser for correct parsing. Unrecognised
    files are appended to the end of the list in the same order they are
    given.
    
    This function has been deprecated as of version 1.8; use:
    cclib.parser.turbomoleparser.Turbomole.sort_input() instead

    Inputs:
      filelist - a list of Turbomole log files needed to be parsed.
    Returns:
      sorted_list - a sorted list of Turbomole files needed for proper parsing.
    """
    warnings.warn(
        "sort_turbomole_outputs() has been deprecated as of v1.8; use: "+ \
        "cclib.parser.turbomoleparser.Turbomole.sort_input() instead")
    return Turbomole.sort_input(fileinputs)

def ccread(
        source: Union[str, typing.IO, FileWrapper, typing.List[Union[str, typing.IO]]],
        *args,
        **kwargs
    ):
    """Attempt to open and read computational chemistry data from a file.

    If the file is not appropriate for cclib parsers, a fallback mechanism
    will try to recognize some common chemistry formats and read those using
    the appropriate bridge such as Open Babel.

    Inputs:
        source - a single logfile, a list of logfiles (for a single job),
                 an input stream, or an URL pointing to a log file.
        *args, **kwargs - arguments and keyword arguments passed to ccopen
    Returns:
        a ccData object containing cclib data attributes
    """
    log = None
    try:
        log = ccopen(source, *args, **kwargs)
        logger = logging.getLogger("cclib")
        if log:
            logger.info("Identified logfile to be in {} format".format(type(log).__name__))

            return log.parse()
        else:
            logger.info('Attempting to use fallback mechanism to read file')
            return fallback(source)
    
    finally:
        if log:
            log.inputfile.close()


def ccopen(
        source: Union[str, typing.IO, FileWrapper, typing.List[Union[str, typing.IO]]],
        *args,
        quiet: bool = False,
        cjson: bool = False,
        **kwargs
    ):
    """Guess the identity of a particular log file and return an instance of it.

    Inputs:
        source - a single logfile, a list of logfiles (for a single job),
                 an input stream, or an URL pointing to a log file.
        *args, **kwargs - arguments and keyword arguments passed to filetype

    Returns:
      one of ADF, DALTON, GAMESS, GAMESS UK, Gaussian, Jaguar,
      Molpro, MOPAC, NWChem, ORCA, Psi3, Psi/Psi4, QChem, CJSON or None
      (if it cannot figure it out or the file does not exist).
    """
    if not isinstance(source, list):
        source = [source]
        
    inputfile = None

    logger = logging.getLogger("cclib")
        
    try:
        # Wrap our input with custom file object.
        inputfile = FileWrapper(*source)
        
        if cjson:
            filetype = readerclasses['cjson']
        
        else:
            # Try and guess the parser we need.
            filetype = guess_filetype(inputfile)
        
            # Reset our position back to 0.
            inputfile.reset()
    
        # If the input file isn't a standard compchem log file, try one of
        # the readers, falling back to Open Babel.
        if not filetype:
            # TODO: This assumes we only got a single file...
            filename = list(inputfile.filenames)[0]
            ext = pathlib.Path(filename).name[1:].lower()
            
            for extension in readerclasses:
                if ext == extension:
                    filetype = readerclasses[extension]
    
        # Proceed to return an instance of the logfile parser only if the filetype
        # could be guessed.
        if filetype:
                return filetype(inputfile, *args, **kwargs)
    
        elif inputfile is not None:
            inputfile.close()
            # Stop us closing twice in the except block.
            inputfile = None
            
        logger.warning(
            "Unable to determine the type of logfile %s, try the fallback mechanism",
            source
        )
            
    except Exception:
        if inputfile is not None:
            inputfile.close()
        
        if not quiet:
            raise
        
        # We're going to swallow this exception if quiet is True.
        # This can hide a lot of errors, so we'll make sure to log it.
        logger.error("Failed to open logfile", exc_info = True)


def fallback(source):
    """Attempt to read standard molecular formats using other libraries.

    Currently this will read XYZ files with OpenBabel, but this can easily
    be extended to other formats and libraries, too.
    """

    if isinstance(source, str):
        ext = os.path.splitext(source)[1][1:].lower()
        if _has_cclib2openbabel:
            # From OB 3.0 onward, Pybel is contained inside the OB module.
            try:
                import openbabel.pybel as pb
            except:
                import pybel as pb
            if ext in pb.informats:
                return cclib2openbabel.readfile(source, ext)
        else:
            # This should be a warning, but warnings are currently disabled by default.
            logging.getLogger("cclib").error(
                "Could not import `openbabel`, fallback mechanism might not work.")


def ccwrite(ccobj, outputtype=None, outputdest=None,
            indices=None, terse=False, returnstr=False,
            *args, **kwargs):
    """Write the parsed data from an outputfile to a standard chemical
    representation.

    Inputs:
        ccobj - Either a job (from ccopen) or a data (from job.parse()) object
        outputtype - The output format (should be a string)
        outputdest - A filename or file object for writing
        indices - One or more indices for extracting specific geometries/etc. (zero-based)
        terse -  This option is currently limited to the cjson/json format. Whether to indent the cjson/json or not
        returnstr - Whether or not to return a string representation.

    The different writers may take additional arguments, which are
    documented in their respective docstrings.

    Returns:
        the string representation of the chemical datatype
          requested, or None.
    """

    # Determine the correct output format.
    outputclass = _determine_output_format(outputtype, outputdest)

    # Is ccobj an job object (unparsed), or is it a ccdata object (parsed)?
    if isinstance(ccobj, logfileparser.Logfile):
        jobfilename = ccobj.filename
        ccdata = ccobj.parse()
    elif isinstance(ccobj, data.ccData):
        jobfilename = None
        ccdata = ccobj
    else:
        raise ValueError

    # If the logfile name has been passed in through kwargs (such as
    # in the ccwrite script), make sure it has precedence.
    if 'jobfilename' in kwargs:
        jobfilename = kwargs['jobfilename']
        # Avoid passing multiple times into the main call.
        del kwargs['jobfilename']

    outputobj = outputclass(ccdata, jobfilename=jobfilename,
                            indices=indices, terse=terse,
                            *args, **kwargs)
    output = outputobj.generate_repr()

    # If outputdest isn't None, write the output to disk.
    if outputdest is not None:
        if isinstance(outputdest, str):
            with open(outputdest, 'w') as outputobj:
                outputobj.write(output)
        elif isinstance(outputdest, io.IOBase):
            outputdest.write(output)
        else:
            raise ValueError
    # If outputdest is None, return a string representation of the output.
    else:
        return output

    if returnstr:
        return output


def _determine_output_format(outputtype, outputdest):
    """
    Determine the correct output format.

    Inputs:
      outputtype - a string corresponding to the file type
      outputdest - a filename string or file handle
    Returns:
      outputclass - the class corresponding to the correct output format
    Raises:
      UnknownOutputFormatError for unsupported file writer extensions
    """

    # Priority for determining the correct output format:
    #  1. outputtype
    #  2. outputdest

    outputclass = None
    # First check outputtype.
    if isinstance(outputtype, str):
        extension = outputtype.lower()
        if extension in writerclasses:
            outputclass = writerclasses[extension]
        else:
            raise UnknownOutputFormatError(extension)
    else:
        # Then checkout outputdest.
        if isinstance(outputdest, str):
            extension = os.path.splitext(outputdest)[1].lower()
        elif isinstance(outputdest, io.IOBase):
            extension = os.path.splitext(outputdest.name)[1].lower()
        else:
            raise UnknownOutputFormatError
        if extension in writerclasses:
            outputclass = writerclasses[extension]
        else:
            raise UnknownOutputFormatError(extension)

    return outputclass


def _check_pandas(found_pandas):
    if not found_pandas:
        raise ImportError("You must install `pandas` to use this function")


def ccframe(ccobjs, *args, **kwargs):
    """Returns a pandas.DataFrame of data attributes parsed by cclib from one
    or more logfiles.

    Inputs:
        ccobjs - an iterable of either cclib jobs (from ccopen) or data (from
        job.parse()) objects

    Returns:
        a pandas.DataFrame
    """
    _check_pandas(_has_pandas)
    logfiles = []
    for ccobj in ccobjs:
        # Is ccobj an job object (unparsed), or is it a ccdata object (parsed)?
        if isinstance(ccobj, logfileparser.Logfile):
            jobfilename = ccobj.filename
            ccdata = ccobj.parse()
        elif isinstance(ccobj, data.ccData):
            jobfilename = None
            ccdata = ccobj
        else:
            raise ValueError

        attributes = ccdata.getattributes()
        attributes.update({
            'jobfilename': jobfilename
        })

        logfiles.append(pd.Series(attributes))
    return pd.DataFrame(logfiles)


del find_package




# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Parser for NWChem output files"""


import itertools
import re

import numpy

from cclib.parser import logfileparser
from cclib.parser import utils


class NWChem(logfileparser.Logfile):
    """An NWChem log file."""

    def __init__(self, *args, **kwargs):
        super().__init__(logname="NWChem", *args, **kwargs)

    def __str__(self):
        """Return a string representation of the object."""
        return f"NWChem log file {self.filename}"

    def __repr__(self):
        """Return a representation of the object."""
        return f'NWChem("{self.filename}")'

    def normalisesym(self, label):
        """NWChem does not require normalizing symmetry labels."""
        return label

    name2element = lambda self, lbl: "".join(itertools.takewhile(str.isalpha, str(lbl)))

    def extract(self, inputfile, line):
        """Extract information from the file object inputfile."""
        # search for No. of atoms     :
        if line[:22] == "          No. of atoms":
            if not hasattr(self, 'natom'):
                natom = int(line[28:])
                self.set_attribute('natom', natom)

        # Extract the version number and the version control information, if
        # it exists.
        if "nwchem branch" in line:
            base_package_version = line.split()[3]
            self.metadata["legacy_package_version"] = base_package_version
            self.metadata["package_version"] = base_package_version
            line = next(inputfile)
            if "nwchem revision" in line:
                revision = line.split()[3].split('-')[-1]
                if revision != "N/A":
                    self.metadata[
                        "package_version"
                    ] = f"{self.metadata['package_version']}+{revision}"

        # This is printed in the input module, so should always be the first coordinates,
        # and contains some basic information we want to parse as well. However, this is not
        # the only place where the coordinates are printed during geometry optimization,
        # since the gradients module has a separate coordinate printout, which happens
        # alongside the coordinate gradients. This geometry printout happens at the
        # beginning of each optimization step only.
        if line.strip() == 'Geometry "geometry" -> ""' or line.strip() == 'Geometry "geometry" -> "geometry"':

            self.skip_lines(inputfile, ['dashes', 'blank', 'units', 'blank', 'header', 'dashes'])

            if not hasattr(self, 'atomcoords'):
                self.atomcoords = []

            line = next(inputfile)
            coords = []
            atomnos = []
            while line.strip():
                # The column labeled 'tag' is usually empty, but I'm not sure whether it can have spaces,
                # so for now assume that it can and that there will be seven columns in that case.
                if len(line.split()) == 6:
                    index, atomname, nuclear, x, y, z = line.split()
                else:
                    index, atomname, tag, nuclear, x, y, z = line.split()
                coords.append(list(map(float, [x, y, z])))
                atomnos.append(int(float(nuclear)))
                line = next(inputfile)

            self.atomcoords.append(coords)

            self.set_attribute('atomnos', atomnos)
            self.set_attribute('natom', len(atomnos))

        if line.strip() == "Atomic Mass":
            self.skip_lines(inputfile, ["d", "b"])
            assert hasattr(self, "atomnos")
            unique_atomnos = set(self.atomnos)
            name2mass = {}
            for _ in range(len(unique_atomnos)):
                tokens = next(inputfile).split()
                mtch = re.search(r"[a-zA-Z]+", tokens[0])
                assert mtch is not None
                name2mass[mtch.group()] = float(tokens[1])
            masses = [
                name2mass[self.table.element[number]]
                for number in self.atomnos
            ]
            self.set_attribute("atommasses", masses)

        if line.strip() == "Symmetry information":
            self.skip_lines(inputfile, ['d', 'b'])
            line = next(inputfile)
            assert line[1:11] == "Group name"
            point_group_full = line.split()[2].lower()
            line = next(inputfile)
            assert line[1:13] == "Group number"
            line = next(inputfile)
            assert line[1:12] == "Group order"
            self.pg_order = int(line.split()[2])
            # TODO It is unclear what class of point group NWChem
            # prints, since it can handle crystallographic groups in
            # some places:
            # http://www.nwchem-sw.org/index.php/Release66:Geometry
            point_group_abelian = point_group_full
            self.metadata['symmetry_detected'] = point_group_full
            self.metadata['symmetry_used'] = point_group_abelian

        # If the geometry is printed in XYZ format, it will have the number of atoms.
        if line[12:31] == "XYZ format geometry":

            self.skip_line(inputfile, 'dashes')
            natom = int(next(inputfile).strip())
            self.set_attribute('natom', natom)

        if line.strip() == "NWChem Geometry Optimization":
            # see cclib/cclib#1057
            self.skip_lines(inputfile, ['d', 'b', 'b'])
            line = next(inputfile)
            if "maximum gradient threshold" not in line:
                self.skip_lines(inputfile, ['b', 'title', 'b', 'b'])
                line = next(inputfile)
            if "no constraints" in line:
                line = next(inputfile)
            assert "maximum gradient threshold" in line
            while line.strip():
                if "maximum gradient threshold" in line:
                    gmax = float(line.split()[-1])
                if "rms gradient threshold" in line:
                    grms = float(line.split()[-1])
                if "maximum cartesian step threshold" in line:
                    xmax = float(line.split()[-1])
                if "rms cartesian step threshold" in line:
                    xrms = float(line.split()[-1])
                line = next(inputfile)

            self.set_attribute('geotargets', [gmax, grms, xmax, xrms])

        # NWChem does not normally print the basis set for each atom, but rather
        # chooses the concise option of printing Gaussian coefficients for each
        # atom type/element only once. Therefore, we need to first parse those
        # coefficients and afterwards build the appropriate gbasis attribute based
        # on that and atom types/elements already parsed (atomnos). However, if atom
        # are given different names (number after element, like H1 and H2), then NWChem
        # generally prints the gaussian parameters for all unique names, like this:
        #
        #                      Basis "ao basis" -> "ao basis" (cartesian)
        #                      -----
        #  O (Oxygen)
        #  ----------
        #            Exponent  Coefficients
        #       -------------- ---------------------------------------------------------
        #  1 S  1.30709320E+02  0.154329
        #  1 S  2.38088610E+01  0.535328
        # (...)
        #
        #  H1 (Hydrogen)
        #  -------------
        #            Exponent  Coefficients
        #       -------------- ---------------------------------------------------------
        #  1 S  3.42525091E+00  0.154329
        # (...)
        #
        #  H2 (Hydrogen)
        #  -------------
        #            Exponent  Coefficients
        #       -------------- ---------------------------------------------------------
        #  1 S  3.42525091E+00  0.154329
        # (...)
        #
        # This current parsing code below assumes all atoms of the same element
        # use the same basis set, but that might not be true, and this will probably
        # need to be considered in the future when such a logfile appears.
        if line.strip() == """Basis "ao basis" -> "ao basis" (cartesian)""":
            self.skip_line(inputfile, 'dashes')
            gbasis_dict = {}
            line = next(inputfile)
            while line.strip():
                atomname = line.split()[0]
                atomelement = self.name2element(atomname)
                gbasis_dict[atomelement] = []
                self.skip_lines(inputfile, ['d', 'labels', 'd'])
                shells = []
                line = next(inputfile)
                while line.strip() and line.split()[0].isdigit():
                    shell = None
                    while line.strip():
                        nshell, type, exp, coeff = line.split()
                        nshell = int(nshell)
                        assert len(shells) == nshell - 1
                        if not shell:
                            shell = (type, [])
                        else:
                            assert shell[0] == type
                        exp = float(exp)
                        coeff = float(coeff)
                        shell[1].append((exp, coeff))
                        line = next(inputfile)
                    shells.append(shell)
                    line = next(inputfile)
                gbasis_dict[atomelement].extend(shells)

            gbasis = []
            for i in range(self.natom):
                atomtype = self.table.element[self.atomnos[i]]
                gbasis.append(gbasis_dict[atomtype])

            self.set_attribute('gbasis', gbasis)

        # Normally the indexes of AOs assigned to specific atoms are also not printed,
        # so we need to infer that. We could do that from the previous section,
        # it might be worthwhile to take numbers from two different places, hence
        # the code below, which builds atombasis based on the number of functions
        # listed in this summary of the AO basis. Similar to previous section, here
        # we assume all atoms of the same element have the same basis sets, but
        # this will probably need to be revised later.

        # The section we can glean info about aonmaes looks like:
        #
        # Summary of "ao basis" -> "ao basis" (cartesian)
        # ------------------------------------------------------------------------------
        #       Tag                 Description            Shells   Functions and Types
        # ---------------- ------------------------------  ------  ---------------------
        # C                           sto-3g                  3        5   2s1p
        # H                           sto-3g                  1        1   1s
        #
        # However, we need to make sure not to match the following entry lines:
        #
        # *  Summary of "ao basis" -> "" (cartesian)
        # *  Summary of allocated global arrays
        #
        # Unfortantely, "ao basis" isn't unique because it can be renamed to anything for
        # later reference: http://www.nwchem-sw.org/index.php/Basis
        # It also appears that we have to handle cartesian vs. spherical

        if line[1:11] == "Summary of":
            match = re.match(r' Summary of "([^\"]*)" -> "([^\"]*)" \((.+)\)', line)

            if match and match.group(1) == match.group(2):

                self.skip_lines(inputfile, ['d', 'title', 'd'])

                self.shells = {}
                self.shells["type"] = match.group(3)

                atombasis_dict = {}

                line = next(inputfile)
                while line.strip():
                    atomname, desc, shells, funcs, types = line.split()
                    atomelement = self.name2element(atomname)
                    self.metadata["basis_set"] = desc

                    self.shells[atomname] = types
                    atombasis_dict[atomelement] = int(funcs)
                    line = next(inputfile)

                last = 0
                atombasis = []
                for atom in self.atomnos:
                    atomelement = self.table.element[atom]
                    nfuncs = atombasis_dict[atomelement]
                    atombasis.append(list(range(last, last+nfuncs)))
                    last = atombasis[-1][-1] + 1

                self.set_attribute('atombasis', atombasis)

        if line.strip() == "Symmetry analysis of basis":
            self.skip_lines(inputfile, ['d', 'b'])
            if not hasattr(self, 'symlabels'):
                self.symlabels = []
            for _ in range(self.pg_order):
                line = next(inputfile)
                self.symlabels.append(self.normalisesym(line.split()[0]))

        # This section contains general parameters for Hartree-Fock calculations,
        # which do not contain the 'General Information' section like most jobs.
        if line.strip() == "NWChem SCF Module":
            # If the calculation doesn't have a title specified, there
            # aren't as many lines to skip here.
            self.skip_lines(inputfile, ['d', 'b', 'b'])
            line = next(inputfile)
            if line.strip():
                self.skip_lines(inputfile, ['b', 'b', 'b'])
            line = next(inputfile)
            while line.strip():
                if line[2:8] == "charge":
                    charge = int(float(line.split()[-1]))
                    self.set_attribute('charge', charge)
                if line[2:13] == "open shells":
                    unpaired = int(line.split()[-1])
                    self.set_attribute('mult', 2*unpaired + 1)
                if line[2:7] == "atoms":
                    natom = int(line.split()[-1])
                    self.set_attribute('natom', natom)
                if line[2:11] == "functions":
                    nfuncs = int(line.split()[-1])
                    self.set_attribute("nbasis", nfuncs)
                line = next(inputfile)

        # This section contains general parameters for DFT calculations, as well as
        # for the many-electron theory module.
        if line.strip() == "General Information":

            if hasattr(self, 'linesearch') and self.linesearch:
                return

            while line.strip():

                if "No. of atoms" in line:
                    self.set_attribute('natom', int(line.split()[-1]))
                if "Charge" in line:
                    self.set_attribute('charge', int(line.split()[-1]))
                if "Spin multiplicity" in line:
                    mult = line.split()[-1]
                    if mult == "singlet":
                        mult = 1
                    self.set_attribute('mult', int(mult))
                if "AO basis - number of function" in line:
                    nfuncs = int(line.split()[-1])
                    self.set_attribute('nbasis', nfuncs)

                # These will be present only in the DFT module.
                if "Convergence on energy requested" in line:
                    self.target_energy = utils.float(line.split()[-1])
                if "Convergence on density requested" in line:
                    self.target_density = utils.float(line.split()[-1])
                if "Convergence on gradient requested" in line:
                    self.target_gradient = utils.float(line.split()[-1])

                line = next(inputfile)

            # set scftargets only in the SCF module.
            scftargetattrs = ("target_energy", "target_density", "target_gradient")
            if self.hasattrs(scftargetattrs):
                self.append_attribute(
                    "scftargets",
                    [getattr(self, attr) for attr in scftargetattrs]
                )
                for attr in scftargetattrs:
                    delattr(self, attr)

        #DFT functional information
        if "XC Information" in line:
            line = next(inputfile)
            line = next(inputfile)
            self.metadata["functional"] = line.split()[0]

        # If the full overlap matrix is printed, it looks like this:
        #
        # global array: Temp Over[1:60,1:60],  handle: -996
        #
        #            1           2           3           4           5           6
        #       ----------- ----------- ----------- ----------- ----------- -----------
        #   1       1.00000     0.24836    -0.00000    -0.00000     0.00000     0.00000
        #   2       0.24836     1.00000     0.00000    -0.00000     0.00000     0.00030
        #   3      -0.00000     0.00000     1.00000     0.00000     0.00000    -0.00014
        # ...
        if "global array: Temp Over[" in line:

            self.set_attribute('nbasis', int(line.split('[')[1].split(',')[0].split(':')[1]))
            self.set_attribute('nmo', int(line.split(']')[0].split(',')[1].split(':')[1]))

            aooverlaps = []
            while len(aooverlaps) < self.nbasis:

                self.skip_line(inputfile, 'blank')

                indices = [int(i) for i in next(inputfile).split()]
                assert indices[0] == len(aooverlaps) + 1

                self.skip_line(inputfile, "dashes")
                data = [next(inputfile).split() for i in range(self.nbasis)]
                indices = [int(d[0]) for d in data]
                assert indices == list(range(1, self.nbasis+1))

                for i in range(1, len(data[0])):
                    vector = [float(d[i]) for d in data]
                    aooverlaps.append(vector)

            self.set_attribute('aooverlaps', aooverlaps)

        if line.strip() in ("The SCF is already converged", "The DFT is already converged"):
            if hasattr(self, 'linesearch') and self.linesearch:
                return
            if hasattr(self, 'scftargets'):
                self.scftargets.append(self.scftargets[-1])
            if hasattr(self, 'scfvalues'):
                self.scfvalues.append(self.scfvalues[-1])

        # The default (only?) SCF algorithm for Hartree-Fock is a preconditioned conjugate
        # gradient method that apparently "always" converges, so this header should reliably
        # signal a start of the SCF cycle. The convergence targets are also printed here.
        if line.strip() == "Quadratically convergent ROHF":

            if hasattr(self, 'linesearch') and self.linesearch:
                return

            while not "Final" in line:

                # Only the norm of the orbital gradient is used to test convergence.
                if line[:22] == " Convergence threshold":
                    target = float(line.split()[-1])
                    if not hasattr(self, "scftargets"):
                        self.scftargets = []
                    self.scftargets.append([target])

                    # This is critical for the stop condition of the section,
                    # because the 'Final Fock-matrix accuracy' is along the way.
                    # It would be prudent to find a more robust stop condition.
                    while list(set(line.strip())) != ["-"]:
                        line = next(inputfile)

                if line.split() == ['iter', 'energy', 'gnorm', 'gmax', 'time']:
                    values = []
                    self.skip_line(inputfile, 'dashes')
                    line = next(inputfile)
                    while line.strip():
                        it, energy, gnorm, gmax, time = line.split()
                        gnorm = utils.float(gnorm)
                        values.append([gnorm])
                        try:
                            line = next(inputfile)
                        # Is this the end of the file for some reason?
                        except StopIteration:
                            self.logger.warning(
                                f"File terminated before end of last SCF! Last gradient norm: {gnorm}"
                            )
                            break
                    if not hasattr(self, 'scfvalues'):
                        self.scfvalues = []
                    self.scfvalues.append(values)

                try:
                    line = next(inputfile)
                except StopIteration:
                    self.logger.warning('File terminated?')
                    break

        # The SCF for DFT does not use the same algorithm as Hartree-Fock, but always
        # seems to use the following format to report SCF convergence:
        #   convergence    iter        energy       DeltaE   RMS-Dens  Diis-err    time
        # ---------------- ----- ----------------- --------- --------- ---------  ------
        # d= 0,ls=0.0,diis     1   -382.2544324446 -8.28D+02  1.42D-02  3.78D-01    23.2
        # d= 0,ls=0.0,diis     2   -382.3017298534 -4.73D-02  6.99D-03  3.82D-02    39.3
        # d= 0,ls=0.0,diis     3   -382.2954343173  6.30D-03  4.21D-03  7.95D-02    55.3
        # ...
        if line.split() == ['convergence', 'iter', 'energy', 'DeltaE', 'RMS-Dens', 'Diis-err', 'time']:

            if hasattr(self, 'linesearch') and self.linesearch:
                return

            self.skip_line(inputfile, 'dashes')
            line = next(inputfile)
            values = []
            while line.strip():

                # Sometimes there are things in between iterations with fewer columns,
                # and we want to skip those lines, most probably. An exception might
                # unrestricted calcualtions, which show extra RMS density and DIIS
                # errors, although it is not clear yet whether these are for the
                # beta orbitals or somethine else. The iterations look like this in that case:
                #   convergence    iter        energy       DeltaE   RMS-Dens  Diis-err    time
                # ---------------- ----- ----------------- --------- --------- ---------  ------
                # d= 0,ls=0.0,diis     1   -382.0243202601 -8.28D+02  7.77D-03  1.04D-01    30.0
                #                                                     7.68D-03  1.02D-01
                # d= 0,ls=0.0,diis     2   -382.0647539758 -4.04D-02  4.64D-03  1.95D-02    59.2
                #                                                     5.39D-03  2.36D-02
                # ...
                if len(line[17:].split()) == 6:
                    iter, energy, deltaE, dens, diis, time = line[17:].split()
                    val_energy = utils.float(deltaE)
                    val_density = utils.float(dens)
                    val_gradient = utils.float(diis)
                    values.append([val_energy, val_density, val_gradient])

                try:
                    line = next(inputfile)
                # Is this the end of the file for some reason?
                except StopIteration:
                    self.logger.warning(
                        f"File terminated before end of last SCF! Last error: {diis}"
                    )
                    break

            if not hasattr(self, 'scfvalues'):
                self.scfvalues = []

            self.scfvalues.append(values)

        # These triggers are supposed to catch the current step in a geometry optimization search
        # and determine whether we are currently in the main (initial) SCF cycle of that step
        # or in the subsequent line search. The step is printed between dashes like this:
        #
        #          --------
        #          Step   0
        #          --------
        #
        # and the summary lines that describe the main SCF cycle for the frsit step look like this:
        #
        #@ Step       Energy      Delta E   Gmax     Grms     Xrms     Xmax   Walltime
        #@ ---- ---------------- -------- -------- -------- -------- -------- --------
        #@    0    -379.76896249  0.0D+00  0.04567  0.01110  0.00000  0.00000      4.2
        #                                                       ok       ok
        #
        # However, for subsequent step the format is a bit different:
        #
        #  Step       Energy      Delta E   Gmax     Grms     Xrms     Xmax   Walltime
        #  ---- ---------------- -------- -------- -------- -------- -------- --------
        #@    2    -379.77794602 -7.4D-05  0.00118  0.00023  0.00440  0.01818     14.8
        #                                              ok
        #
        # There is also a summary of the line search (which we don't use now), like this:
        #
        # Line search:
        #     step= 1.00 grad=-1.8D-05 hess= 8.9D-06 energy=   -379.777955 mode=accept
        # new step= 1.00                   predicted energy=   -379.777955
        #
        if line[10:14] == "Step":
            self.geostep = int(line.split()[-1])
            self.skip_line(inputfile, 'dashes')
            self.linesearch = False
        if line[0] == "@" and line.split()[1] == "Step":
            at_and_dashes = next(inputfile)
            line = next(inputfile)
            tokens = line.split()
            assert int(tokens[1]) == self.geostep == 0
            gmax = float(tokens[4])
            grms = float(tokens[5])
            xrms = float(tokens[6])
            xmax = float(tokens[7])
            self.append_attribute("geovalues", [gmax, grms, xmax, xrms])
            self.linesearch = True
        if line[2:6] == "Step":
            self.skip_line(inputfile, 'dashes')
            line = next(inputfile)
            tokens = line.split()
            assert int(tokens[1]) == self.geostep
            gmax = float(tokens[4])
            grms = float(tokens[5])
            xrms = float(tokens[6])
            xmax = float(tokens[7])
            if hasattr(self, 'linesearch') and self.linesearch:
                return
            self.append_attribute("geovalues", [gmax, grms, xmax, xrms])
            self.linesearch = True

        # There is a clear message when the geometry optimization has converged:
        #
        #      ----------------------
        #      Optimization converged
        #      ----------------------
        #
        if line.strip() == "Optimization converged":
            self.skip_line(inputfile, 'dashes')
            if not hasattr(self, 'optdone'):
                self.optdone = []
            self.optdone.append(len(self.geovalues) - 1)

        if "Failed to converge" in line and hasattr(self, 'geovalues'):
            if not hasattr(self, 'optdone'):
                self.optdone = []

        # extract the theoretical method
        if "Total SCF energy" in line:
            self.metadata["methods"].append("HF")
        if "Total DFT energy" in line:
            self.metadata["methods"].append("DFT")

        # The line containing the final SCF energy seems to be always identifiable like this.
        if "Total SCF energy" in line or "Total DFT energy" in line:

            # NWChem often does a line search during geometry optimization steps, reporting
            # the SCF information but not the coordinates (which are not necessarily 'intermediate'
            # since the step size can become smaller). We want to skip these SCF cycles,
            # unless the coordinates can also be extracted (possibly from the gradients?).
            if hasattr(self, 'linesearch') and self.linesearch:
                return

            if not hasattr(self, "scfenergies"):
                self.scfenergies = []
            energy = float(line.split()[-1])
            energy = utils.convertor(energy, "hartree", "eV")
            self.scfenergies.append(energy)

        if "Dispersion correction" in line:
            dispersion = utils.convertor(float(line.split()[-1]), "hartree", "eV")
            self.append_attribute("dispersionenergies", dispersion)

        # The final MO orbitals are printed in a simple list, but apparently not for
        # DFT calcs, and often this list does not contain all MOs, so make sure to
        # parse them from the MO analysis below if possible. This section will be like this:
        #
        #       Symmetry analysis of molecular orbitals - final
        #       -----------------------------------------------
        #
        #  Numbering of irreducible representations:
        #
        #     1 ag          2 au          3 bg          4 bu
        #
        #  Orbital symmetries:
        #
        #     1 bu          2 ag          3 bu          4 ag          5 bu
        #     6 ag          7 bu          8 ag          9 bu         10 ag
        # ...
        if line.strip() == "Symmetry analysis of molecular orbitals - final":

            self.skip_lines(inputfile, ['d', 'b', 'numbering', 'b', 'reps', 'b', 'syms', 'b'])

            if not hasattr(self, 'mosyms'):
                self.mosyms = [[None]*self.nbasis]
            line = next(inputfile)
            while line.strip():
                ncols = len(line.split())
                assert ncols % 2 == 0
                for i in range(ncols//2):
                    index = int(line.split()[i*2]) - 1
                    sym = line.split()[i*2+1]
                    sym = sym[0].upper() + sym[1:]
                    if self.mosyms[0][index]:
                        if self.mosyms[0][index] != sym:
                            self.logger.warning(
                                f"Symmetry of MO {int(index + 1)} has changed"
                            )
                    self.mosyms[0][index] = sym
                line = next(inputfile)

        # The same format is used for HF and DFT molecular orbital analysis. We want to parse
        # the MO energies from this section, although it is printed already before this with
        # less precision (might be useful to parse that if this is not available). Also, this
        # section contains coefficients for the leading AO contributions, so it might also
        # be useful to parse and use those values if the full vectors are not printed.
        #
        # The block looks something like this (two separate alpha/beta blocks in the unrestricted case):
        #
        #                       ROHF Final Molecular Orbital Analysis
        #                       -------------------------------------
        #
        # Vector    1  Occ=2.000000D+00  E=-1.104059D+01  Symmetry=bu
        #              MO Center=  1.4D-17,  0.0D+00, -6.5D-37, r^2= 2.1D+00
        #   Bfn.  Coefficient  Atom+Function         Bfn.  Coefficient  Atom+Function
        #  ----- ------------  ---------------      ----- ------------  ---------------
        #     1      0.701483   1 C  s                 6     -0.701483   2 C  s
        #
        # Vector    2  Occ=2.000000D+00  E=-1.104052D+01  Symmetry=ag
        # ...
        # Vector   12  Occ=2.000000D+00  E=-1.020253D+00  Symmetry=bu
        #              MO Center= -1.4D-17, -5.6D-17,  2.9D-34, r^2= 7.9D+00
        #   Bfn.  Coefficient  Atom+Function         Bfn.  Coefficient  Atom+Function
        #  ----- ------------  ---------------      ----- ------------  ---------------
        #    36     -0.298699  11 C  s                41      0.298699  12 C  s
        #     2      0.270804   1 C  s                 7     -0.270804   2 C  s
        #    48     -0.213655  15 C  s                53      0.213655  16 C  s
        # ...
        #
        if "Final" in line and "Molecular Orbital Analysis" in line:

            # Unrestricted jobs have two such blocks, for alpha and beta orbitals, and
            # we need to keep track of which one we're parsing (always alpha in restricted case).
            unrestricted = ("Alpha" in line) or ("Beta" in line)
            alphabeta = int("Beta" in line)

            self.skip_lines(inputfile, ['dashes', 'blank'])

            nvectors = []
            mooccnos = []
            energies = []
            symmetries = [None]*self.nbasis
            line = next(inputfile)
            while line[:7] == " Vector":

                # Note: the vector count starts from 1 in NWChem.
                nvector = int(line[7:12])
                nvectors.append(nvector)

                # A nonzero occupancy for SCF jobs means the orbital is occupied.
                mooccno = int(utils.float(line[18:30]))
                mooccnos.append(mooccno)

                # If the printout does not start from the first MO, assume None for all previous orbitals.
                if len(energies) == 0 and nvector > 1:
                    for i in range(1, nvector):
                        energies.append(None)

                energy = utils.float(line[34:47])
                energy = utils.convertor(energy, "hartree", "eV")
                energies.append(energy)

                # When symmetry is not used, this part of the line is missing.
                if line[47:58].strip() == "Symmetry=":
                    sym = line[58:].strip()
                    sym = sym[0].upper() + sym[1:]
                    symmetries[nvector-1] = sym

                line = next(inputfile)
                if "MO Center" in line:
                    line = next(inputfile)
                if "Bfn." in line:
                    line = next(inputfile)
                if "-----" in line:
                    line = next(inputfile)
                while line.strip():
                    line = next(inputfile)
                line = next(inputfile)

            self.set_attribute('nmo', nvector)

            if not hasattr(self, 'moenergies') or (len(self.moenergies) > alphabeta):
                self.moenergies = []
            self.moenergies.append(energies)

            if not hasattr(self, 'mosyms') or (len(self.mosyms) > alphabeta):
                self.mosyms = []
            self.mosyms.append(symmetries)

            if not hasattr(self, 'homos') or (len(self.homos) > alphabeta):
                self.homos = []
            nvector_index = mooccnos.index(0) - 1
            if nvector_index > -1:
                self.homos.append(nvectors[nvector_index] - 1)
            else:
                self.homos.append(-1)
            # If this was a restricted open-shell calculation, append
            # to HOMOs twice since only one Molecular Orbital Analysis
            # section is in the output file.
            if (not unrestricted) and (1 in mooccnos):
                nvector_index = mooccnos.index(1) - 1
                if nvector_index > -1:
                    self.homos.append(nvectors[nvector_index] - 1)
                else:
                    self.homos.append(-1)

        # This is where the full MO vectors are printed, but a special
        # directive is needed for it in the `scf` or `dft` block:
        #   print "final vectors" "final vectors analysis"
        # which gives:
        #
        #                                 Final MO vectors
        #                                 ----------------
        #
        #
        # global array: alpha evecs[1:60,1:60],  handle: -995
        #
        #            1           2           3           4           5           6
        #       ----------- ----------- ----------- ----------- ----------- -----------
        #   1      -0.69930    -0.69930    -0.02746    -0.02769    -0.00313    -0.02871
        #   2      -0.03156    -0.03135     0.00410     0.00406     0.00078     0.00816
        #   3       0.00002    -0.00003     0.00067     0.00065    -0.00526    -0.00120
        # ...
        #
        if line.strip() == "Final MO vectors":

            if not hasattr(self, 'mocoeffs'):
                self.mocoeffs = []

            self.skip_lines(inputfile, ['d', 'b', 'b'])

            # The columns are MOs, rows AOs, but that's and educated guess since no
            # atom information is printed alongside the indices. This next line gives
            # the dimensions, which we can check. if set before this. Also, this line
            # specifies whether we are dealing with alpha or beta vectors.
            array_info = next(inputfile)
            while ("global array" in array_info):
                alphabeta = int(line.split()[2] == "beta")
                size = array_info.split('[')[1].split(']')[0]
                nbasis = int(size.split(',')[0].split(':')[1])
                nmo = int(size.split(',')[1].split(':')[1])
                self.set_attribute('nbasis', nbasis)
                self.set_attribute('nmo', nmo)

                self.skip_line(inputfile, 'blank')
                mocoeffs = []
                while len(mocoeffs) < self.nmo:
                    nmos = list(map(int, next(inputfile).split()))
                    assert len(mocoeffs) == nmos[0] - 1
                    for n in nmos:
                        mocoeffs.append([])
                    self.skip_line(inputfile, 'dashes')
                    for nb in range(nbasis):
                        line = next(inputfile)
                        index = int(line.split()[0])
                        assert index == nb+1
                        coefficients = list(map(float, line.split()[1:]))
                        assert len(coefficients) == len(nmos)
                        for i, c in enumerate(coefficients):
                            mocoeffs[nmos[i]-1].append(c)
                    self.skip_line(inputfile, 'blank')
                self.mocoeffs.append(mocoeffs)

                array_info = next(inputfile)

        # For Hartree-Fock, the atomic Mulliken charges are typically printed like this:
        #
        #  Mulliken analysis of the total density
        #  --------------------------------------
        #
        #    Atom       Charge   Shell Charges
        # -----------   ------   -------------------------------------------------------
        #    1 C    6     6.00   1.99  1.14  2.87
        #    2 C    6     6.00   1.99  1.14  2.87
        # ...
        if line.strip() == "Mulliken analysis of the total density":

            if not hasattr(self, "atomcharges"):
                self.atomcharges = {}

            self.skip_lines(inputfile, ['d', 'b', 'header', 'd'])

            charges = []
            line = next(inputfile)
            while line.strip():
                index, atomname, nuclear, atom = line.split()[:4]
                shells = line.split()[4:]
                charges.append(float(nuclear)-float(atom))
                line = next(inputfile)
            self.atomcharges['mulliken'] = charges

        # Note the 'overlap population' as printed in the Mulliken population analysis
        # is not the same thing as the 'overlap matrix'. In fact, it is the overlap matrix
        # multiplied elementwise times the density matrix.
        #
        #          ----------------------------
        #          Mulliken population analysis
        #          ----------------------------
        #
        #          ----- Total      overlap population -----
        #
        #                               1              2              3              4              5              6              7
        #
        #    1   1 C  s            2.0694818227  -0.0535883400  -0.0000000000  -0.0000000000  -0.0000000000  -0.0000000000   0.0000039991
        #    2   1 C  s           -0.0535883400   0.8281341291   0.0000000000  -0.0000000000   0.0000000000   0.0000039991  -0.0009906747
        # ...
        #
        # DFT does not seem to print the separate listing of Mulliken charges
        # by default, but they are printed by this modules later on. They are also print
        # for Hartree-Fock runs, though, so in that case make sure they are consistent.
        if line.strip() == "Mulliken population analysis":

            self.skip_lines(inputfile, ['d', 'b', 'total_overlap_population', 'b'])

            overlaps = []
            line = next(inputfile)
            while all([c.isdigit() for c in line.split()]):

                # There is always a line with the MO indices printed in thie block.
                indices = [int(i)-1 for i in line.split()]
                for i in indices:
                    overlaps.append([])

                # There is usually a blank line after the MO indices, but
                # there are exceptions, so check if line is blank first.
                line = next(inputfile)
                if not line.strip():
                    line = next(inputfile)

                # Now we can iterate or atomic orbitals.
                for nao in range(self.nbasis):
                    data = list(map(float, line.split()[4:]))
                    for i, d in enumerate(data):
                        overlaps[indices[i]].append(d)
                    line = next(inputfile)

                line = next(inputfile)

            # This header should be printed later, before the charges are print, which of course
            # are just sums of the overlaps and could be calculated. But we just go ahead and
            # parse them, make sure they're consistent with previously parsed values and
            # use these since they are more precise (previous precision could have been just 0.01).
            while "Total      gross population on atoms" not in line:
                line = next(inputfile)
            self.skip_line(inputfile, 'blank')
            charges = []
            for i in range(self.natom):
                line = next(inputfile)
                iatom, element, ncharge, epop = line.split()
                iatom = int(iatom)
                ncharge = float(ncharge)
                epop = float(epop)
                assert iatom == (i+1)
                charges.append(ncharge-epop)

            if not hasattr(self, 'atomcharges'):
                self.atomcharges = {}
            if "mulliken" in self.atomcharges:
                assert max(self.atomcharges['mulliken'] - numpy.array(charges)) < 0.01
            # This is going to be higher precision than "Mulliken analysis of
            # the total density".
            self.atomcharges['mulliken'] = charges

        # NWChem prints the dipole moment in atomic units first, and we could just fast forward
        # to the values in Debye, which are also printed. But we can also just convert them
        # right away and so parse a little bit less. Note how the reference point is print
        # here within the block nicely, as it is for all moment later.
        #
        #          -------------
        #          Dipole Moment
        #          -------------
        #
        # Center of charge (in au) is the expansion point
        #         X =       0.0000000 Y =       0.0000000 Z =       0.0000000
        #
        #   Dipole moment        0.0000000000 Debye(s)
        #             DMX        0.0000000000 DMXEFC        0.0000000000
        #             DMY        0.0000000000 DMYEFC        0.0000000000
        #             DMZ       -0.0000000000 DMZEFC        0.0000000000
        #
        # ...
        #
        if line.strip() == "Dipole Moment":

            self.skip_lines(inputfile, ['d', 'b'])

            reference_comment = next(inputfile)
            assert "(in au)" in reference_comment
            reference = next(inputfile).split()
            self.reference = [reference[-7], reference[-4], reference[-1]]
            self.reference = numpy.array([float(x) for x in self.reference])
            self.reference = utils.convertor(self.reference, 'bohr', 'Angstrom')

            self.skip_line(inputfile, 'blank')

            magnitude = next(inputfile)
            assert magnitude.split()[-1] == "A.U."

            dipole = []
            for i in range(3):
                line = next(inputfile)
                dipole.append(float(line.split()[1]))

            dipole = utils.convertor(numpy.array(dipole), "ebohr", "Debye")

            if not hasattr(self, 'moments'):
                self.moments = [self.reference, dipole]
            else:
                self.moments[1] == dipole

        # The quadrupole moment is pretty straightforward to parse. There are several
        # blocks printed, and the first one called 'second moments' contains the raw
        # moments, and later traceless values are printed. The moments, however, are
        # not in lexicographical order, so we need to sort them. Also, the first block
        # is in atomic units, so remember to convert to Buckinghams along the way.
        #
        #          -----------------
        #          Quadrupole Moment
        #          -----------------
        #
        # Center of charge (in au) is the expansion point
        #         X =       0.0000000 Y =       0.0000000 Z =       0.0000000
        #
        # < R**2 > = ********** a.u.  ( 1 a.u. = 0.280023 10**(-16) cm**2 )
        # ( also called diamagnetic susceptibility )
        #
        #   Second moments in atomic units
        #
        #   Component  Electronic+nuclear     Point charges             Total
        #  --------------------------------------------------------------------------
        #      XX          -38.3608511210          0.0000000000        -38.3608511210
        #      YY          -39.0055467347          0.0000000000        -39.0055467347
        # ...
        #
        if line.strip() == "Quadrupole Moment":

            self.skip_lines(inputfile, ['d', 'b'])

            reference_comment = next(inputfile)
            assert "(in au)" in reference_comment
            reference = next(inputfile).split()
            self.reference = [reference[-7], reference[-4], reference[-1]]
            self.reference = numpy.array([float(x) for x in self.reference])
            self.reference = utils.convertor(self.reference, 'bohr', 'Angstrom')

            self.skip_lines(inputfile, ['b', 'units', 'susc', 'b'])

            line = next(inputfile)
            assert line.strip() == "Second moments in atomic units"

            self.skip_lines(inputfile, ['b', 'header', 'd'])

            # Parse into a dictionary and then sort by the component key.
            quadrupole = {}
            for i in range(6):
                line = next(inputfile)
                quadrupole[line.split()[0]] = float(line.split()[-1])
            lex = sorted(quadrupole.keys())
            quadrupole = [quadrupole[key] for key in lex]

            quadrupole = utils.convertor(numpy.array(quadrupole), "ebohr2", "Buckingham")

            # The checking of potential previous values if a bit more involved here,
            # because it turns out NWChem has separate keywords for dipole, quadrupole
            # and octupole output. So, it is perfectly possible to print the quadrupole
            # and not the dipole... if that is the case set the former to None and
            # issue a warning. Also, a regression has been added to cover this case.
            if not hasattr(self, 'moments') or len(self.moments) < 2:
                self.logger.warning("Found quadrupole moments but no previous dipole")
                self.moments = [self.reference, None, quadrupole]
            else:
                if len(self.moments) == 2:
                    self.moments.append(quadrupole)
                else:
                    assert self.moments[2] == quadrupole

        # The octupole moment is analogous to the quadrupole, but there are more components
        # and the checking of previously parsed dipole and quadrupole moments is more involved,
        # with a corresponding test also added to regressions.
        #
        #          ---------------
        #          Octupole Moment
        #          ---------------
        #
        # Center of charge (in au) is the expansion point
        #         X =       0.0000000 Y =       0.0000000 Z =       0.0000000
        #
        #   Third moments in atomic units
        #
        #   Component  Electronic+nuclear     Point charges             Total
        #  --------------------------------------------------------------------------
        #      XXX          -0.0000000000          0.0000000000         -0.0000000000
        #      YYY          -0.0000000000          0.0000000000         -0.0000000000
        # ...
        #
        if line.strip() == "Octupole Moment":

            self.skip_lines(inputfile, ['d', 'b'])

            reference_comment = next(inputfile)
            assert "(in au)" in reference_comment
            reference = next(inputfile).split()
            self.reference = [reference[-7], reference[-4], reference[-1]]
            self.reference = numpy.array([float(x) for x in self.reference])
            self.reference = utils.convertor(self.reference, 'bohr', 'Angstrom')

            self.skip_line(inputfile, 'blank')

            line = next(inputfile)
            assert line.strip() == "Third moments in atomic units"

            self.skip_lines(inputfile, ['b', 'header', 'd'])

            octupole = {}
            for i in range(10):
                line = next(inputfile)
                octupole[line.split()[0]] = float(line.split()[-1])
            lex = sorted(octupole.keys())
            octupole = [octupole[key] for key in lex]

            octupole = utils.convertor(numpy.array(octupole), "ebohr3", "Debye.ang2")

            if not hasattr(self, 'moments') or len(self.moments) < 2:
                self.logger.warning("Found octupole moments but no previous dipole or quadrupole moments")
                self.moments = [self.reference, None, None, octupole]
            elif len(self.moments) == 2:
                self.logger.warning("Found octupole moments but no previous quadrupole moments")
                self.moments.append(None)
                self.moments.append(octupole)
            else:
                if len(self.moments) == 3:
                    self.moments.append(octupole)
                else:
                    assert self.moments[3] == octupole

        if "Total MP2 energy" in line:
            self.metadata["methods"].append("MP2")
            mpenerg = float(line.split()[-1])
            if not hasattr(self, "mpenergies"):
                self.mpenergies = []
            self.mpenergies.append([])
            self.mpenergies[-1].append(utils.convertor(mpenerg, "hartree", "eV"))

        if line.strip() == "NWChem Extensible Many-Electron Theory Module":
            ccenergies = []
            while "Parallel integral file used" not in line:
                line = next(inputfile)
                if "CCSD total energy / hartree" in line or "total CCSD energy:" in line:
                    self.metadata["methods"].append("CCSD")
                    ccenergies.append(float(line.split()[-1]))
                if "CCSD(T) total energy / hartree" in line:
                    self.metadata["methods"].append("CCSD(T)")
                    ccenergies.append(float(line.split()[-1]))
            if ccenergies:
                self.append_attribute(
                    "ccenergies", utils.convertor(ccenergies[-1], "hartree", "eV")
                )

        # Static and dynamic polarizability.
        if "Linear Response polarizability / au" in line:
            if not hasattr(self, "polarizabilities"):
                self.polarizabilities = []
            polarizability = []
            line = next(inputfile)
            assert line.split()[0] == "Frequency"
            line = next(inputfile)
            assert line.split()[0] == "Wavelength"
            self.skip_lines(inputfile, ['coordinates', 'd'])
            for _ in range(3):
                line = next(inputfile)
                polarizability.append(line.split()[1:])
            self.polarizabilities.append(numpy.array(polarizability))

        if line[:18] == ' Total times  cpu:':
            self.metadata['success'] = True

        if line.strip() == "NWChem QMD Module":
            self.is_BOMD = True

        # Born-Oppenheimer molecular dynamics (BOMD): time.
        if "QMD Run Information" in line:
            self.skip_line(inputfile, 'd')
            line = next(inputfile)
            assert "Time elapsed (fs)" in line
            time = float(line.split()[4])
            self.append_attribute('time', time)

        if "ENERGY GRADIENTS" in line:
            self.skip_lines(inputfile, ['b', 'atom coordinates gradient', 'xyzxyz'])
            line = next(inputfile)
            atomcoords_step = []
            grad_step = []
            while line.strip():
                tokens = line.split()
                assert len(tokens) == 8
                atomcoords_step.append([float(c) for c in tokens[2:5]])
                grad_step.append([float(x) for x in tokens[5:8]])
                line = next(inputfile)
            self.append_attribute("grads", grad_step)
            # BOMD: geometry coordinates when `print low`.
            if self.is_BOMD:
                self.append_attribute("atomcoords", atomcoords_step)

        if "Atom information" in line:
            self.skip_lines(inputfile, ["atom # X Y Z mass", "d"])
            masses = []
            for _ in range(self.natom):
                line = next(inputfile)
                masses.append(utils.float(line[61:74]))
            self.set_attribute("atommasses", masses)

        # Section with atomic coordinates and masses (in higher precision than
        # "Atomic Mass") printed in vibrational calculations.
        if "Atom information" in line:
            atommasses = []
            self.skip_lines(inputfile, ["header", "d"])
            line = next(inputfile)
            while set(line.strip()) != {"-"}:
                atommasses.append(utils.float(line.split()[-1]))
                line = next(inputfile)
            self.set_attribute("atommasses", atommasses)

        if "MASS-WEIGHTED NUCLEAR HESSIAN" in line:
            nelem = 3 * self.natom
            hessian = numpy.empty((nelem, nelem))
            rc = 1
            cc = 1
            self.skip_lines(inputfile, ["d"])
            while cc < nelem:
                lines = self.skip_lines(inputfile, ["b", "b", "column numbers", "d"])
                cc_prev = cc
                cc = int(lines[2].split()[-1])
                while rc < nelem:
                    line = next(inputfile)
                    tokens = line.split()
                    rc = int(tokens[0])
                    vals = [utils.float(x) for x in tokens[1:]]
                    cstart = cc_prev - 1
                    cend = cstart + len(vals)
                    hessian[rc - 1, cstart:cend] = vals
                rc = cc + 1
            self.set_attribute("hessian", utils.symmetrize(hessian))

        # Extract Thermochemistry in au (Hartree)
        #
        # have to deal with :
        # Temperature                      =   298.15K
        # frequency scaling parameter      =   1.0000
        # Zero-Point correction to Energy  =  259.352 kcal/mol  (  0.413304 au)
        # Thermal correction to Energy     =  275.666 kcal/mol  (  0.439302 au)
        # Thermal correction to Enthalpy   =  276.258 kcal/mol  (  0.440246 au)
        # Total Entropy                    =  176.764 cal/mol-K
        # - Translational                =   44.169 cal/mol-K (mol. weight = 448.1245)
        # - Rotational                   =   37.018 cal/mol-K (symmetry #  =        1)
        # - Vibrational                  =   95.577 cal/mol-K
        # Cv (constant volume heat capacity) =  103.675 cal/mol-K
        # - Translational                  =    2.979 cal/mol-K
        # - Rotational                     =    2.979 cal/mol-K
        # - Vibrational                    =   97.716 cal/mol-K
        if line[1:12] == "Temperature":
            self.set_attribute("temperature", utils.float(line.split()[2][:-1]))
        if line[1:28] == "frequency scaling parameter":
            self.set_attribute("pressure", utils.float(line.split()[4]))
        if line[1:31] == "Thermal correction to Enthalpy" and hasattr(self, "scfenergies"):
            self.set_attribute(
                "enthalpy",
                utils.float(line.split()[8])
                + utils.convertor(self.scfenergies[-1], "eV", "hartree"),
            )
        if line[1:32] == "Zero-Point correction to Energy":
            self.set_attribute("zpve", float(line.split()[8]))
        if line[1:29] == "Thermal correction to Energy" and hasattr(self, "scfenergies"):
            self.set_attribute(
                "electronic_thermal_energy",
                utils.float(line.split()[8])
                + utils.convertor(self.scfenergies[-1], "eV", "hartree"),
            )
        if line[1:14] == "Total Entropy":
            self.set_attribute(
                "entropy",
                utils.convertor(1e-3 * utils.float(line.split()[3]), "kcal/mol", "hartree"),
            )

        if line.strip() == "(Projected Frequencies expressed in cm-1)":
            # We can't look for "NORMAL MODE EIGENVECTORS IN CARTESIAN
            # COORDINATES" since it also appears for unprojected normal modes.
            vibfreqs = []
            vibdisps = numpy.empty(shape=(3 * self.natom, self.natom, 3))
            mode_idx = 0
            while mode_idx < 3 * self.natom - 1:
                lines = self.skip_lines(
                    inputfile,
                    ["b", "one-based mode index", "b"]
                )
                tokens = next(inputfile).split()
                nmodes_block = len(tokens) - 1
                mode_idx_end = mode_idx + nmodes_block
                assert tokens[0] == "P.Frequency"
                vibfreqs.extend([float(x) for x in tokens[1:]])
                self.skip_line(inputfile, "b")
                for idx_atom in range(self.natom):
                    for idx_coord in range(3):
                        tokens = next(inputfile).split()[1:]
                        vibdisps[mode_idx:mode_idx_end, idx_atom, idx_coord] = tokens
                mode_idx = mode_idx_end
            self.set_attribute("vibfreqs", vibfreqs)
            self.set_attribute("vibdisps", vibdisps)

        # extract vibrational frequencies (in cm-1)
        if line.strip() == "Normal Eigenvalue ||           Projected Infra Red Intensities":
            self.skip_lines(inputfile, ["units", "d"])  # units, dashes
            line = next(inputfile)  # first line of data
            vibfreqs = []
            while set(line.strip()[:-1]) != {"-"}:
                vibfreqs.append(float(line.split()[1]))
                self.append_attribute("vibirs", float(line.split()[5]))
                line = next(inputfile)  # next line
            self.set_attribute("vibfreqs", vibfreqs)
        # NWChem TD-DFT excited states transitions
        #
        # Have to deal with :
        # ----------------------------------------------------------------------------
        # Root   1 singlet a              0.105782828 a.u.                2.8785 eV
        # ----------------------------------------------------------------------------
        #    Transition Moments    X -1.88278   Y -0.46346   Z -0.05660
        #    Transition Moments   XX -5.63612  XY  4.57009  XZ -0.38291
        #    Transition Moments   YY  6.48024  YZ -1.50109  ZZ -0.17430
        #    Dipole Oscillator Strength                    0.2653650650
        #    Electric Quadrupole                           0.0000003789
        #    Magnetic Dipole                               0.0000001767
        #    Total Oscillator Strength                     0.2653656206
        #
        #    Occ.  117  a   ---  Virt.  118  a    0.98676 X
        #    Occ.  117  a   ---  Virt.  118  a   -0.08960 Y
        #    Occ.  117  a   ---  Virt.  119  a    0.08235 X
        # ----------------------------------------------------------------------------
        # Root   2 singlet a              0.127858653 a.u.                3.4792 eV
        # ----------------------------------------------------------------------------
        #    Transition Moments    X -0.02031   Y  0.11238   Z -0.09893
        #    Transition Moments   XX -0.23065  XY -0.35697  XZ -0.11250
        #    Transition Moments   YY  0.16402  YZ -0.01716  ZZ  0.16705
        #    Dipole Oscillator Strength                    0.0019460560
        #    Electric Quadrupole                           0.0000000021
        #    Magnetic Dipole                               0.0000002301
        #    Total Oscillator Strength                     0.0019462882
        #
        #    Occ.  110  a   ---  Virt.  118  a   -0.05918 X
        #    Occ.  110  a   ---  Virt.  119  a   -0.06022 X
        #    Occ.  110  a   ---  Virt.  124  a    0.05962 X
        #    Occ.  114  a   ---  Virt.  118  a    0.87840 X
        #    Occ.  114  a   ---  Virt.  119  a   -0.12213 X
        #    Occ.  114  a   ---  Virt.  123  a    0.07120 X
        #    Occ.  114  a   ---  Virt.  124  a   -0.05022 X
        #    Occ.  114  a   ---  Virt.  125  a    0.06104 X
        #    Occ.  114  a   ---  Virt.  126  a    0.05065 X
        #    Occ.  115  a   ---  Virt.  118  a    0.12907 X
        #    Occ.  116  a   ---  Virt.  118  a   -0.40137 X

        if line[:6] == "  Root":
            self.append_attribute(
                "etenergies", utils.convertor(utils.float(line.split()[-2]), "eV", "wavenumber")
            )
            self.append_attribute("etsyms", str.join(" ", line.split()[2:-4]))
        
            self.skip_lines(inputfile, ["dashes"])
            line = next(inputfile)
            if "Spin forbidden" not in line:
                # find Dipole Oscillator Strength
                while not ("Dipole Oscillator Strength" in line):
                    line = next(inputfile)
                etoscs = utils.float(line.split()[-1])
                # in case of magnetic contribution replace, replace Dipole Oscillator Strength with Total Oscillator Strength
                while not (line.find("Occ.") >= 0):
                    if "Total Oscillator Strength" in line:
                        etoscs = utils.float(line.split()[-1])
                    line = next(inputfile)
                self.append_attribute("etoscs", etoscs)
                CIScontrib = []
                while line.find("Occ.") >= 0:
                    if len(line.split()) == 9:  # restricted
                        _, occ, _, _, _, virt, _, coef, direction = line.split()
                        type1 = "alpha"
                        type2 = "alpha"
                    else:  # unrestricted: len(line.split()) should be 11
                        _, occ, type1, _, _, _, virt, type2, _, coef, direction = line.split()
                    occ = int(occ) - 1  # subtract 1 so that it is an index into moenergies
                    virt = int(virt) - 1  # subtract 1 so that it is an index into moenergies
                    coef = utils.float(coef)
                    if direction == "Y":
                        # imaginary or negative excitation (denoted Y)
                        tmp = virt
                        virt = occ
                        occ = tmp
                        tmp = type1
                        type1 = type2
                        type2 = tmp
                    frommoindex = 0  # For restricted or alpha unrestricted
                    if type1 == "beta":
                        frommoindex = 1
                    tomoindex = 0  # For restricted or alpha unrestricted
                    if type2 == "beta":
                        tomoindex = 1
                    CIScontrib.append([(occ, frommoindex), (virt, tomoindex), coef])
                    line = next(inputfile)
                self.append_attribute("etsecs", CIScontrib)
            else:
                self.append_attribute("etoscs", 0.0)

    def before_parsing(self):
        """NWChem-specific routines performed before parsing a file.
        """

        # The only reason we need this identifier is if `print low` is
        # set in the input file, which we assume is likely for a BOMD
        # trajectory. This will enable parsing coordinates from the
        # 'DFT ENERGY GRADIENTS' section.
        self.is_BOMD = False

    def after_parsing(self):
        """NWChem-specific routines for after parsing a file.
        """
        super(NWChem, self).after_parsing()

        # Expand self.shells into a proper aonames attribute.
        #
        # setup a few necessary things, including a regular expression
        # for matching the shells
        table = utils.PeriodicTable()
        elements = [table.element[x] for x in self.atomnos]
        pattern = re.compile(r"(\ds)+(\dp)*(\dd)*(\df)*(\dg)*")

        labels = {}
        labels['s'] = ["%iS"]
        labels['p'] = ["%iPX", "%iPY", "%iPZ"]
        if self.shells['type'] == 'spherical':
            labels['d'] = ['%iD-2', '%iD-1', '%iD0', '%iD1', '%iD2']
            labels['f'] = ['%iF-3', '%iF-2', '%iF-1', '%iF0',
                           '%iF1', '%iF2', '%iF3']
            labels['g'] = ['%iG-4', '%iG-3', '%iG-2', '%iG-1', '%iG0',
                           '%iG1', '%iG2', '%iG3', '%iG4']
        elif self.shells['type'] == 'cartesian':
            labels['d'] = ['%iDXX', '%iDXY', '%iDXZ',
                           '%iDYY', '%iDYZ',
                           '%iDZZ']
            labels['f'] = ['%iFXXX', '%iFXXY', '%iFXXZ',
                           '%iFXYY', '%iFXYZ', '%iFXZZ',
                           '%iFYYY', '%iFYYZ', '%iFYZZ',
                           '%iFZZZ']
            labels['g'] = ['%iGXXXX', '%iGXXXY', '%iGXXXZ',
                           '%iGXXYY', '%iGXXYZ', '%iGXXZZ',
                           '%iGXYYY', '%iGXYYZ', '%iGXYZZ',
                           '%iGXZZZ', '%iGYYYY', '%iGYYYZ',
                           '%iGYYZZ', '%iGYZZZ', '%iGZZZZ']
        else:
            self.logger.warning("Found a non-standard aoname representation type.")
            return

        # now actually build aonames
        # involves expanding 2s1p into appropriate types

        self.aonames = []
        for i, element in enumerate(elements):
            try:
                shell_text = self.shells[element]
            except KeyError:
                del self.aonames
                msg = "Cannot determine aonames for at least one atom."
                self.logger.warning(msg)
                break

            prefix = f"{element}{int(i + 1)}_"  # (e.g. C1_)

            matches = pattern.match(shell_text)
            for j, group in enumerate(matches.groups()):
                if group is None:
                    continue

                count = int(group[:-1])
                label = group[-1]

                for k in range(count):
                    temp = [x % (j + k + 1) for x in labels[label]]
                    self.aonames.extend([prefix + x for x in temp])

        # If we parsed a BOMD trajectory, the first two parsed
        # geometries are identical, and all from the second onward are
        # in Bohr. Delete the first one and perform the unit
        # conversion.
        if self.is_BOMD:
            self.atomcoords = utils.convertor(numpy.asarray(self.atomcoords)[1:, ...],
                                              'bohr', 'Angstrom')

        # Versions >= 7.0 print an extra "General Information" section
        # containing scftargets before entering "NWChem DFT Gradient Module".
        # Since originally it only appeared before entering SCF and not also
        # afterwards, keep every other starting from the first appearance.
        if self.hasattrs(("grads", "scftargets")) \
           and len(self.scftargets) == (2 * len(self.grads)):
            self.scftargets = self.scftargets[::2]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Test logfiles with vibration output in cclib"""

import os
import unittest

from skip import skipForParser, skipForLogfile


__filedir__ = os.path.realpath(os.path.dirname(__file__))


class GenericIRTest(unittest.TestCase):
    """Generic vibrational frequency unittest"""

    # Unit tests should normally give this value for the largest IR intensity.
    max_IR_intensity = 100

    # Unit tests may give these values for the largest force constant and reduced mass, respectively.
    max_force_constant = 10.0
    max_reduced_mass = 6.9

    # reference zero-point correction from Gaussian 16 dvb_ir.out
    zpve = 0.1771

    def setUp(self) -> None:
        """Initialize the number of vibrational frequencies on a per molecule basis"""
        self.numvib = 3*len(self.data.atomnos) - 6

    def testbasics(self) -> None:
        """Are basic attributes correct?"""
        assert self.data.natom == 20

    @skipForLogfile("FChk/basicGaussian09", "not printed in older versions than 16")
    @skipForLogfile("FChk/basicQChem5.4", "not printed")
    def testvibdisps(self) -> None:
        """Are the dimensions of vibdisps consistent with numvib x N x 3"""
        assert len(self.data.vibfreqs) == self.numvib
        assert self.data.vibdisps.shape == (self.numvib, len(self.data.atomnos), 3)

    @skipForLogfile("FChk/basicGaussian09", "not printed in older versions than 16")
    @skipForLogfile("FChk/basicQChem5.4", "not printed")
    def testlengths(self) -> None:
        """Are the lengths of vibfreqs and vibirs (and if present, vibsyms, vibfconnsts and vibrmasses) correct?"""
        assert len(self.data.vibfreqs) == self.numvib
        if hasattr(self.data, 'vibirs'):
            assert len(self.data.vibirs) == self.numvib
        if hasattr(self.data, 'vibsyms'):
            assert len(self.data.vibsyms) == self.numvib
        if hasattr(self.data, 'vibfconsts'):
            assert len(self.data.vibfconsts) == self.numvib
        if hasattr(self.data, 'vibrmasses'):
            assert len(self.data.vibrmasses) == self.numvib

    @skipForLogfile("FChk/basicGaussian09", "not printed in older versions than 16")
    @skipForLogfile("FChk/basicQChem5.4", "not printed")
    def testfreqval(self) -> None:
        """Is the highest freq value 3630 +/- 200 wavenumber?"""
        assert abs(max(self.data.vibfreqs) - 3630) < 200

    @skipForParser('Psi4', 'Psi cannot print IR intensities')
    @skipForLogfile("FChk/basicGaussian09", "not printed in older versions than 16")
    @skipForLogfile("FChk/basicQChem5.4", "not printed")
    def testirintens(self) -> None:
        """Is the maximum IR intensity 100 +/- 10 km/mol?"""
        assert abs(max(self.data.vibirs) - self.max_IR_intensity) < 10

    @skipForParser('ADF', 'ADF cannot print force constants')
    @skipForParser('DALTON', 'DALTON cannot print force constants')
    @skipForParser('GAMESS', 'GAMESS-US cannot print force constants')
    @skipForParser('GAMESSUK', 'GAMESS-UK cannot print force constants')
    @skipForParser('Molcas', 'Molcas cannot print force constants')
    @skipForParser('Molpro', 'Molpro cannot print force constants')
    @skipForParser('NWChem', 'Not implemented for this parser')
    @skipForParser('ORCA', 'ORCA cannot print force constants')
    @skipForParser('Turbomole', 'Turbomole cannot print force constants')
    @skipForLogfile('Jaguar/Jaguar4.2', 'Data file does not contain force constants')
    @skipForLogfile('Psi4/Psi4-1.0', 'Data file contains vibrational info with cartesian coordinates')
    @skipForLogfile("FChk/basicGaussian09", "not printed in older versions than 16")
    @skipForLogfile("FChk/basicQChem5.4", "not printed")
    def testvibfconsts(self) -> None:
        """Is the maximum force constant 10. +/- 0.1 mDyn/angstrom?"""
        assert abs(max(self.data.vibfconsts) - self.max_force_constant) < 0.1

    @skipForParser('ADF', 'ADF cannot print reduced masses')
    @skipForParser('DALTON', 'DALTON cannot print reduced masses')
    @skipForParser('GAMESSUK', 'GAMESSUK cannot print reduced masses')
    @skipForParser('Molpro', 'Molpro cannot print reduced masses')
    @skipForParser('NWChem', 'Not implemented for this parser')
    @skipForParser('ORCA', 'ORCA cannot print reduced masses')
    @skipForLogfile('GAMESS/PCGAMESS', 'Data file does not contain reduced masses')
    @skipForLogfile('Psi4/Psi4-1.0', 'Data file does not contain reduced masses')
    @skipForLogfile("FChk/basicGaussian09", "not printed in older versions than 16")
    @skipForLogfile("FChk/basicQChem5.4", "not printed")
    def testvibrmasses(self) -> None:
        """Is the maximum reduced mass 6.9 +/- 0.1 daltons?"""
        assert abs(max(self.data.vibrmasses) - self.max_reduced_mass) < 0.1

    @skipForParser("FChk", "not printed")
    @skipForParser('Psi3', 'not implemented yet')
    def testzeropointcorrection(self) -> None:
        """Is the zero-point correction correct?"""
        assert abs(self.data.zpve - self.zpve) < 1.0e-3

    @skipForParser('ADF', 'not implemented yet')
    @skipForParser('GAMESSUK', 'not implemented yet')
    @skipForParser('Gaussian', 'not implemented yet')
    @skipForParser('Jaguar', 'not implemented yet')
    @skipForParser('Molcas', 'not implemented yet')
    @skipForParser('Molpro', 'not implemented yet')
    @skipForParser('ORCA', 'not implemented yet')
    @skipForParser('Psi4', 'not implemented yet')
    @skipForLogfile('QChem/basicQChem5.4/dvb_ir.out', 'needs to be rerun with print level turned up')
    @skipForParser('Turbomole', 'not implemented yet')
    def testhessian(self):
        """Are the dimensions of the molecular Hessian correct?"""
        assert self.data.hessian.shape == (3 * self.data.natom, 3 * self.data.natom)


class ADFIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    # ???
    def testzeropointcorrection(self) -> None:
        """Is the zero-point correction correct?"""
        assert abs(self.data.zpve - self.zpve) < 1.0e-2


class FireflyIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    max_IR_intensity = 135
    # ???
    zpve = 0.1935


class GaussianIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    @skipForParser("FChk", "not printed")
    def testvibsyms(self) -> None:
        """Is the length of vibsyms correct?"""
        assert len(self.data.vibsyms) == self.numvib

    @skipForParser("FChk", "not printed")
    def testzeropointcorrection(self) -> None:
        # reference zero-point correction from dvb_ir.out
        zpve = 0.1771
        """Is the zero-point correction correct?"""
        assert abs(self.data.zpve - zpve) < 0.001

    entropy_places = 6
    enthalpy_places = 3
    freeenergy_places = 3

    @skipForParser("FChk", "not printed")
    def testtemperature(self) -> None:
        """Is the temperature 298.15 K?"""
        assert round(abs(298.15 - self.data.temperature), 7) == 0

    @skipForParser("FChk", "not printed")
    def testpressure(self) -> None:
        """Is the pressure 1 atm?"""
        assert round(abs(1 - self.data.pressure), 7) == 0

    @skipForParser("FChk", "not printed")
    def testentropy(self) -> None:
         """Is the entropy reasonable"""
         assert round(abs(0.0001462623335480945 - self.data.entropy), self.entropy_places) == 0

    @skipForParser("FChk", "not printed")
    def testenthalpy(self) -> None:
         """Is the enthalpy reasonable"""
         assert round(abs(-382.12130688525264 - self.data.enthalpy), self.enthalpy_places) == 0

    @skipForParser("FChk", "not printed")
    def testfreeenergy(self) -> None:
         """Is the freeenergy reasonable"""
         assert round(abs(-382.164915 - self.data.freeenergy), self.freeenergy_places) == 0

    @skipForParser("FChk", "not printed")
    def testfreeenergyconsistency(self) -> None:
        """Does G = H - TS hold"""
        assert round(
            abs(
                self.data.enthalpy - self.data.temperature * self.data.entropy -self .data.freeenergy
            ),
            self.freeenergy_places
        ) == 0

class JaguarIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    # Jagur outputs vibrational info with cartesian coordinates
    max_force_constant = 3.7
    max_reduced_mass = 2.3

    def testvibsyms(self) -> None:
        """Is the length of vibsyms correct?"""
        assert len(self.data.vibsyms) == self.numvib


class MolcasIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    max_IR_intensity = 65
    zpve = 0.1783

    entropy_places = 6
    enthalpy_places = 3
    freeenergy_places = 3

    def testtemperature(self) -> None:
        """Is the temperature 298.15 K?"""
        assert round(abs(298.15 - self.data.temperature), 7) == 0

    def testpressure(self) -> None:
        """Is the pressure 1 atm?"""
        assert round(abs(1 - self.data.pressure), 7) == 0

    def testentropy(self) -> None:
         """Is the entropy reasonable"""
         assert round(abs(0.00013403320476271246 - self.data.entropy), self.entropy_places) == 0

    def testenthalpy(self) -> None:
         """Is the enthalpy reasonable"""
         assert round(abs(-382.11385 - self .data.enthalpy), self.enthalpy_places) == 0

    def testfreeenergy(self) -> None:
         """Is the freeenergy reasonable"""
         assert round(abs(-382.153812 - self .data.freeenergy), self.freeenergy_places) == 0

    def testfreeenergyconsistency(self) -> None:
        """Does G = H - TS hold"""
        assert round(
            abs(
                self.data.enthalpy - self.data.temperature * self.data.entropy - self.data.freeenergy
            ),
            self.freeenergy_places
        ) == 0

class NWChemIRTest(GenericIRTest):
    """Generic imaginary vibrational frequency unittest"""

    def setUp(self) -> None:
        """Initialize the number of vibrational frequencies on a per molecule basis"""
        self.numvib = 3*len(self.data.atomnos)


class OrcaIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    # ORCA has a bug in the intensities for version < 4.0
    max_IR_intensity = 215
    zpve = 0.1921

    enthalpy_places = 3
    entropy_places = 6
    freeenergy_places = 3

    def testtemperature(self) -> None:
        """Is the temperature 298.15 K?"""
        assert round(abs(298.15 - self.data.temperature), 7) == 0

    def testpressure(self) -> None:
        """Is the pressure 1 atm?"""
        assert round(abs(1 - self.data.pressure), 7) == 0

    def testenthalpy(self) -> None:
         """Is the enthalpy reasonable"""
         assert round(abs(-381.85224835 - self.data.enthalpy), self.enthalpy_places) == 0

    def testentropy(self) -> None:
         """Is the entropy reasonable"""
         assert round(abs(0.00012080325339594164 - self.data.entropy), self.entropy_places) == 0

    def testfreeenergy(self) -> None:
         """Is the freeenergy reasonable"""
         assert round(abs(-381.88826585 - self.data.freeenergy), self.freeenergy_places) == 0

    def testfreeenergyconsistency(self) -> None:
        """Does G = H - TS hold"""
        assert round(
            abs(
                self.data.enthalpy - self.data.temperature * self.data.entropy - self.data.freeenergy
            ),
            self.freeenergy_places
        ) == 0


class QChemIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    enthalpy_places = 3
    entropy_places = 6
    freeenergy_places = 3

    @skipForParser("FChk", "not printed")
    def testtemperature(self) -> None:
        """Is the temperature 298.15 K?"""
        assert 298.15 == self.data.temperature

    @skipForParser("FChk", "not printed")
    def testpressure(self) -> None:
        """Is the pressure 1 atm?"""
        assert round(abs(1 - self.data.pressure), 7) == 0

    @skipForParser("FChk", "not printed")
    def testenthalpy(self) -> None:
         """Is the enthalpy reasonable"""
         assert round(abs(0.1871270552135131 - self.data.enthalpy), self.enthalpy_places) == 0

    @skipForParser("FChk", "not printed")
    def testentropy(self) -> None:
         """Is the entropy reasonable"""
         assert round(abs(0.00014667348271900577 - self.data.entropy), self.entropy_places) == 0

    @skipForParser("FChk", "not printed")
    def testfreeenergy(self) -> None:
         """Is the freeenergy reasonable"""
         assert round(abs(0.14339635634084155 - self.data.freeenergy), self.freeenergy_places) == 0

    # Molecular mass of DVB in mD.
    molecularmass = 130078.25

    @skipForParser("FChk", "not printed")
    def testatommasses(self) -> None:
        """Do the atom masses sum up to the molecular mass (130078.25+-0.1mD)?"""
        mm = 1000 * sum(self.data.atommasses)
        assert abs(mm - 130078.25) < 0.1, f"Molecule mass: {mm:f} not 130078 +- 0.1mD"

    def testhessian(self) -> None:
        """Do the frequencies from the Hessian match the printed frequencies?"""

    @skipForParser("FChk", "not printed")
    def testfreeenergyconsistency(self) -> None:
        """Does G = H - TS hold"""
        assert round(
            abs(
                self.data.enthalpy - self.data.temperature * self.data.entropy - self.data.freeenergy
            ),
            self.freeenergy_places
        ) == 0


class GamessIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""
    # Molecular mass of DVB in mD.
    molecularmass = 130078.25
    enthalpy_places = 3
    entropy_places = 6
    freeenergy_places = 3

    def testatommasses(self) -> None:
        """Do the atom masses sum up to the molecular mass (130078.25+-0.1mD)?"""
        mm = 1000 * sum(self.data.atommasses)
        assert abs(mm - 130078.25) < 0.1, f"Molecule mass: {mm:f} not 130078 +- 0.1mD"

    def testtemperature(self) -> None:
        """Is the temperature 298.15 K?"""
        assert round(abs(298.15 - self.data.temperature), 7) == 0

    def testpressure(self) -> None:
        """Is the pressure 1 atm?"""
        assert round(abs(1 - self.data.pressure), 7) == 0

    def testenthalpy(self) -> None:
         """Is the enthalpy reasonable"""
         assert round(abs(-381.86372805188300 - self.data.enthalpy), self.enthalpy_places) == 0

    def testentropy(self) -> None:
         """Is the entropy reasonable"""
         assert round(abs(0.00014875961938 - self.data.entropy), self.entropy_places) == 0

    def testfreeenergy(self) -> None:
         """Is the freeenergy reasonable"""
         assert round(abs(-381.90808120060200 - self.data.freeenergy), self.freeenergy_places) == 0

    def testfreeenergyconsistency(self) -> None:
        """Does G = H - TS hold"""
        assert round(
            abs(
                self.data.enthalpy - self.data.temperature * self.data.entropy - self.data.freeenergy
            ),
            self.freeenergy_places
        ) == 0


class Psi4IRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    # RHF is used for Psi4 IR test data instead of B3LYP
    max_force_constant = 9.37
    zpve = 0.1917


class TurbomoleIRTest(GenericIRTest):
    """Customized vibrational frequency unittest"""

    # ???
    zpve = 0.1725


class GenericIRimgTest(unittest.TestCase):
    """Generic imaginary vibrational frequency unittest"""

    def setUp(self) -> None:
        """Initialize the number of vibrational frequencies on a per molecule basis"""
        self.numvib = 3*len(self.data.atomnos) - 6

    def testvibdisps(self) -> None:
        """Are the dimensions of vibdisps consistent with numvib x N x 3"""
        assert self.data.vibdisps.shape == (self.numvib, len(self.data.atomnos), 3)

    def testlengths(self) -> None:
        """Are the lengths of vibfreqs and vibirs correct?"""
        assert len(self.data.vibfreqs) == self.numvib
        assert len(self.data.vibirs) == self.numvib

    def testfreqval(self) -> None:
        """Is the lowest freq value negative?"""
        assert self.data.vibfreqs[0] < 0

##    def testmaxvibdisps(self):
##        """What is the maximum value of displacement for a H vs a C?"""
##        Cvibdisps = compress(self.data.atomnos==6, self.data.vibdisps, 1)
##        Hvibdisps = compress(self.data.atomnos==1, self.data.vibdisps, 1)
##        self.assertEqual(max(abs(Cvibdisps).flat), 1.0)


class GenericRamanTest(unittest.TestCase):
    """Generic Raman unittest"""

    # This value is in amu.
    max_raman_intensity = 575

    def setUp(self) -> None:
        """Initialize the number of vibrational frequencies on a per molecule basis"""
        self.numvib = 3*len(self.data.atomnos) - 6

    def testlengths(self) -> None:
        """Is the length of vibramans correct?"""
        assert len(self.data.vibramans) == self.numvib

    # The tolerance for this number has been increased, since ORCA
    # failed to make it inside +/-5, but it would be nice in the future
    # to determine is it's not too much work whether this is due to
    # algorithmic differences, or to differences in the input basis set
    # or coordinates. The first would be OK, but in the second case the
    # unit test jobs should be made more comparable. With cclib, we first
    # of all want to succeed in parsing, but would also like to remain
    # as comparable between programs as possible (for these tests).
    # Note also that this value is adjusted for Gaussian and DALTON - why?
    def testramanintens(self) -> None:
        """Is the maximum Raman intensity correct?"""
        assert abs(max(self.data.vibramans) - self.max_raman_intensity) < 8

        # We used to test this, but it seems to vary wildly between
        # programs... perhaps we could use it if we knew why...
        #self.assertInside(self.data.vibramans[1], 2.6872, 0.0001)

    def testvibdisps(self) -> None:
        """Is the length and value of vibdisps correct?"""
        assert hasattr(self.data, "vibdisps")
        assert len(self.data.vibdisps) == 54


class DALTONRamanTest(GenericRamanTest):
    """Customized Raman unittest"""

    max_raman_intensity = 745


class GaussianRamanTest(GenericRamanTest):
    """Customized Raman unittest"""

    max_raman_intensity = 1066


class OrcaRamanTest(GenericRamanTest):
    """Customized Raman unittest"""

    max_raman_intensity = 1045


class QChemRamanTest(GenericRamanTest):
    """Customized Raman unittest"""

    max_raman_intensity = 588

if __name__=="__main__":

    import sys
    sys.path.insert(1, os.path.join(__filedir__, ".."))

    from test_data import DataSuite
    suite = DataSuite(['vib'])
    suite.testall()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Contains parsers for all supported programs"""


# These import statements are added for the convenience of users...
# Rather than having to type:
#         from cclib.parser.gaussianparser import Gaussian
# they can use:
#         from cclib.parser import Gaussian

from cclib.parser.adfparser import ADF
from cclib.parser.daltonparser import DALTON
from cclib.parser.fchkparser import FChk
from cclib.parser.gamessparser import GAMESS
from cclib.parser.gamessdatparser import GAMESSDAT
from cclib.parser.gamessukparser import GAMESSUK
from cclib.parser.gaussianparser import Gaussian
from cclib.parser.jaguarparser import Jaguar
from cclib.parser.molcasparser import Molcas
from cclib.parser.molproparser import Molpro
from cclib.parser.mopacparser import MOPAC
from cclib.parser.nboparser import NBO
from cclib.parser.nwchemparser import NWChem
from cclib.parser.orcaparser import ORCA
from cclib.parser.psi3parser import Psi3
from cclib.parser.psi4parser import Psi4
from cclib.parser.qchemparser import QChem
from cclib.parser.turbomoleparser import Turbomole

from cclib.parser.data import ccData

# This allows users to type:
#         from cclib.parser import ccopen
from cclib.io.ccio import ccopen



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Test logfiles with (non)linear response output in cclib"""

import os
import unittest

import numpy

from skip import skipForParser

__filedir__ = os.path.realpath(os.path.dirname(__file__))


class GenericPolarTest(unittest.TestCase):
    """Generic static polarizability unittest"""

    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testshape(self):
        """Is the dimension of the polarizability tensor 3 x 3?"""
        assert len(self.data.polarizabilities) == 1
        assert self.data.polarizabilities[0].shape == (3, 3)


class ReferencePolarTest(GenericPolarTest):
    """Customized static polarizability unittest"""

    # Reference values are from DALTON2015/Trp_polar_abalnr.out
    isotropic = 74.12424
    principal_components = [30.29431523, 91.5361917, 100.54220307]
    isotropic_delta = 0.01
    principal_components_delta = 0.01

    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testisotropic(self):
        """Is the isotropic polarizability (average of the diagonal elements)
        +/- 0.01 from a reference?
        """
        isotropic = numpy.average(numpy.diag(self.data.polarizabilities[0]))
        assert abs(isotropic-self.isotropic) < self.isotropic_delta

    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testprincomponents(self):
        """Are each of the principal components (eigenvalues) of the
        polarizability tensor +/- 0.01 from a reference?
        """
        principal_components = numpy.linalg.eigvalsh(self.data.polarizabilities[0])
        for c in range(3):
            assert abs(principal_components[c]-self.principal_components[c]) < \
                                   self.principal_components_delta


if __name__=="__main__":

    import sys
    sys.path.insert(1, os.path.join(__filedir__, ".."))

    from test_data import DataSuite
    suite = DataSuite(['Polar'])
    suite.testall()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Test single point time-dependent logfiles in cclib"""

import os
import unittest

import numpy

from skip import skipForParser
from skip import skipForLogfile


__filedir__ = os.path.realpath(os.path.dirname(__file__))


class GenericTDTest(unittest.TestCase):
    """Generic time-dependent HF/DFT unittest"""

    number = 5
    expected_l_max = 41000
    expected_f_max = 0.67
    symmetries = [
            "Singlet-Bu",
            "Singlet-Bu",
            "Singlet-Ag",
            "Singlet-Bu",
            "Singlet-Ag",
        ]
    sumofsec = 1.0
    method = "TD-DFT"
    
    @skipForParser('ADF', 'excited_states_method not yet implemented')
    @skipForParser('DALTON', 'excited_states_method not yet implemented')
    @skipForParser('FChk', 'excited_states_method not yet implemented')
    @skipForParser('GAMESS', 'excited_states_method not yet implemented')
    @skipForParser('GAMESSUK', 'excited_states_method not yet implemented')
    @skipForParser('Jaguar', 'excited_states_method not yet implemented')
    @skipForParser('NWChem', 'excited_states_method not yet implemented')
    @skipForParser('QChem', 'excited_states_method not yet implemented')
    def testmetadata(self):
        """Did we parse an excited states method?"""
        assert self.data.metadata['excited_states_method'] == self.method

    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForLogfile('Turbomole/basicTurbomole7.4/CO_cc2_TD_trip', 'Oscillator strengths are not available for Turbomole triplets using ricc2 but are required for testenergies()')
    def testenergies(self):
        """Is the l_max reasonable?"""

        assert len(self.data.etenergies) == self.number

        # Note that if all oscillator strengths are zero (like for triplets)
        # then this will simply pick out the first energy.
        idx_lambdamax = numpy.argmax(self.data.etoscs)
        assert abs(self.data.etenergies[idx_lambdamax] - self.expected_l_max) < 5000

    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForLogfile("Turbomole/basicTurbomole7.4/CO_cc2_TD_trip", "Oscillator strengths are not available for triplets with Turbomole's ricc2")
    def testoscs(self):
        """Is the maximum of etoscs in the right range?"""
        assert len(self.data.etoscs) == self.number
        assert abs(max(self.data.etoscs) - self.expected_f_max) < 0.1

    @skipForParser('FChk','The parser is still being developed so we skip this test')
    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForLogfile("Gaussian/basicGaussian16/dvb_eomccsd.log", "Transitions are not yet parsed for EOM-CCSD")
    def testsecs(self):
        """Is the sum of etsecs close to 1?"""
        assert len(self.data.etsecs) == self.number
        lowestEtrans = self.data.etsecs[numpy.argmin(self.data.etenergies)]
        sumofsec = sum([z*z for (x, y, z) in lowestEtrans])
        assert abs(sumofsec - self.sumofsec) < 0.16

    @skipForParser('FChk', 'This is true for calculations without symmetry, but not with?')
    @skipForParser('DALTON', 'This is true for calculations without symmetry, but not with?')
    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForLogfile("Gaussian/basicGaussian16/dvb_eomccsd.log", "Transitions are not yet parsed for EOM-CCSD")
    def testsecs_transition(self):
        """Is the lowest E transition from the HOMO or to the LUMO?"""
        lowestEtrans = self.data.etsecs[numpy.argmin(self.data.etenergies)]
        t = list(reversed(sorted([(c*c, s, e) for (s, e, c) in lowestEtrans])))
        assert t[0][1][0] == self.data.homos[0] or \
            t[0][2][0] == self.data.homos[0] + 1

    @skipForParser('Molcas','The parser is still being developed so we skip this test')    
    @skipForLogfile("FChk/basicQChem5.4", "etsyms are not yet implemented")
    @skipForLogfile("ORCA/basicORCA5.0/dvb_adc2.log", "etsyms are not available for this method") 
    @skipForLogfile("ORCA/basicORCA5.0/dvb_eom_ccsd.log", "etsyms are not available for this method") 
    @skipForLogfile("ORCA/basicORCA5.0/dvb_pno_eom_ccsd.log", "etsyms are not available for this method") 
    def testsymsnumber(self):
        """Is the length of etsyms correct?"""
        assert len(self.data.etsyms) == self.number
        
    
    @skipForParser('ADF', 'etrotats are not yet implemented')
    @skipForParser('DALTON', 'etsyms are not yet implemented')
    @skipForParser('FChk', 'etsyms are not yet implemented')
    @skipForParser('GAMESS', 'etsyms are not yet implemented')
    @skipForParser('GAMESSUK', 'etsyms are not yet implemented')
    @skipForParser('Jaguar', 'etsyms are not yet implemented')
    @skipForParser('NWChem', 'etsyms are not yet implemented')
    @skipForParser('QChem', 'etsyms are not yet implemented')
    @skipForLogfile("ORCA/basicORCA4.2", "etsyms are only available in ORCA >= 5.0") 
    @skipForLogfile("ORCA/basicORCA4.1", "etsyms are only available in ORCA >= 5.0") 
    @skipForLogfile("Gaussian/basicGaussian09", "symmetry is missing for this log file")
    @skipForLogfile("FChk/basicQChem5.4", "etsyms are not yet implemented")
    def testsyms(self):
        """Are the values of etsyms correct?"""
        assert self.data.etsyms == self.symmetries

    @skipForParser('ADF', 'etrotats are not yet implemented')
    @skipForParser('DALTON', 'etrotats are not yet implemented')
    @skipForParser('FChk', 'etrotats are not yet implemented')
    @skipForParser('GAMESS', 'etrotats are not yet implemented')
    @skipForParser('GAMESSUK', 'etrotats are not yet implemented')
    @skipForParser('Jaguar', 'etrotats are not yet implemented')
    @skipForParser('NWChem', 'etrotats are not yet implemented')
    @skipForParser('QChem', 'Q-Chem cannot calculate rotatory strengths')
    @skipForLogfile("FChk/basicQChem5.4", "Q-Chem cannot calculate rotatory strengths")
    @skipForLogfile("FChk/basicGaussian09", "etrotats are not available in fchk, only the main logfile")
    @skipForLogfile("FChk/basicGaussian16", "etrotats are not available in fchk, only the main logfile")
    @skipForLogfile("Turbomole/basicTurbomole7.4/CO_cc2_TD", "Rotatory strengths are not currently available for ricc2")
    @skipForLogfile("Turbomole/basicTurbomole7.4/CO_adc2_TD", "Rotatory strengths are not currently available for ricc2")
    def testrotatsnumber(self):
        """Is the length of etrotats correct?"""
        assert len(self.data.etrotats) == self.number
    
    @skipForParser('ADF', 'optstate is not yet implemented')
    @skipForParser('DALTON', 'optstate are not yet implemented')
    @skipForParser('FChk', 'optstate are not yet implemented')
    @skipForParser('GAMESS', 'optstate are not yet implemented')
    @skipForParser('GAMESSUK', 'optstate are not yet implemented')
    @skipForParser('Jaguar', 'optstate are not yet implemented')
    @skipForParser('NWChem', 'optstate are not yet implemented')
    @skipForParser('ORCA', 'optstate are not yet implemented')
    @skipForParser('QChem', 'optstate are not yet implemented')
    @skipForParser('Turbomole', 'optstate are not yet implemented')
    def testoptstate(self):
        # All our examples have a default state-of-interest of 1 (index 0).
        assert self.data.metadata['opt_state'] == 0

class ADFTDDFTTest(GenericTDTest):
    """Customized time-dependent DFT unittest"""
    number = 5

    def testsecs(self):
        """Is the sum of etsecs close to 1?"""
        assert len(self.data.etsecs) == self.number
        lowestEtrans = self.data.etsecs[1]

        #ADF squares the etsecs
        sumofsec = sum([z for (x, y, z) in lowestEtrans])
        assert abs(sumofsec - 1.0) < 0.16


class DALTONTDTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""

    number = 20


class GaussianTDDFTTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""

    expected_l_max = 48000

    def testetdipsshape(self):
        """Is the shape of etdips correct?"""
        assert numpy.shape(self.data.etdips) == (self.number, 3)

    def testetveldipsshape(self):
        """Is the shape of etveldips correct?"""
        assert numpy.shape(self.data.etveldips) == (self.number, 3)

    def testetmagdipsshape(self):
        """Is the shape of etmagdips correct?"""
        assert numpy.shape(self.data.etmagdips) == (self.number, 3)

class GAMESSUSTDDFTTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""
    number = 10


class JaguarTDDFTTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""

    expected_l_max = 48000
    expected_f_max = 1.2

class OrcaTDDFTTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""

    number = 10
    expected_l_max = 48000
    symmetries = [
            "Triplet-Bu",
            "Triplet-Ag",
            "Triplet-Bu",
            "Triplet-Bu",
            "Triplet-Bu",
            "Singlet-Bu",
            "Singlet-Bu",
            "Singlet-Ag",
            "Singlet-Bu",
            "Singlet-Ag",
        ]
    method = "TDA"

    def testoscs(self):
        """Is the maximum of etoscs in the right range?"""
        assert len(self.data.etoscs) == self.number
        assert abs(max(self.data.etoscs) - 1.17) < 0.01

class QChemTDDFTTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""

    number = 10
    expected_l_max = 48000
    expected_f_max = 0.9


class GenericTDDFTtrpTest(GenericTDTest):
    """Generic time-dependent HF/DFT (triplet) unittest"""

    number = 5
    expected_l_max = 24500

    def testoscs(self):
        """Triplet excitations should be disallowed."""
        assert len(self.data.etoscs) == self.number
        assert abs(max(self.data.etoscs)) < 0.01


class OrcaROCISTest(GenericTDTest):
    """Customized test for ROCIS"""
    number = 4
    expected_l_max = 2316970.8
    expected_f_max = 0.015
    # per 1085, no VELOCITY DIPOLE MOMENTS are parsed
    n_spectra = 7
    
    # Do we want to parse ROCIS as its own method?
    method = "CIS"

    def testTransprop(self):
        """Check the number of spectra parsed"""
        assert len(self.data.transprop) == self.n_spectra
        tddft_length = 'ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS'
        assert tddft_length in self.data.transprop

    def testsymsnumber(self):
        """ORCA ROCIS has no symmetry"""
        pass

    def testsecs(self):
        """ROCIS does not form singly excited configurations (secs)"""
        pass

    def testsecs_transition(self):
        """ROCIS does not form singly excited configurations (secs)"""
        pass

    def testrotatsnumber(self):
        """ROCIS does not calculate rotatory strengths"""
        pass
    
    def testsyms(self):
        """ROCIS does not show symmetries"""
        pass


class OrcaROCIS40Test(OrcaROCISTest):
    """Customized test for ROCIS"""
    # In ORCA 4.0, an additional spectrum ("COMBINED ELECTRIC DIPOLE +
    # MAGNETIC DIPOLE + ELECTRIC QUADRUPOLE SPECTRUM (Origin Independent,
    # Length Representation)") was present that is not in ORCA 4.1.
    # Changed via 1085. VELOCITY DIPOLE MOMENTS are not parsed.
    n_spectra = 8


class TurbomoleTDTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""
    
    number = 10
    expected_l_max = 91432
    expected_f_max = 0.19
    symmetries = ["Singlet-A"] * 10
    
    @skipForLogfile('Turbomole/basicTurbomole7.4/CO_cc2_TD', 'There are no dipole moments in ricc2')
    def testetmagdipsshape(self):
        """Is the shape of etmagdips correct?"""
        assert numpy.shape(self.data.etmagdips) == (self.number, 3)

class TurbomoleTDADC2Test(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""
    
    number = 10
    expected_l_max = 136329
    expected_f_max = 0.8
    symmetries = ["Singlet-A"] * 10
    method = "ADC(2)"
    
class TurbomoleTDCC2Test(TurbomoleTDTest):
    """Customized time-dependent HF/DFT unittest"""
    
    method = "CC2"

class TurbomoleTDTripTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""
    
    number = 10
    expected_l_max = 51530
    expected_f_max = 0.84
    symmetries = ["Triplet-A"] * 10
    method = "RPA"
        
class TurbomoleTDCC2TripTest(GenericTDTest):
    """Customized time-dependent HF/DFT unittest"""
    # This test is for triplets with ricc2, which does not support oscillator strengths.
    
    number = 10
    symmetries = ["Triplet-A"] * 10
    method = "CC2"

    def testenergies(self):
        """Is the l_max reasonable?"""
        assert len(self.data.etenergies) == self.number

class OrcaETPostHFTest(GenericTDTest):
    """Tests for post-HF excited states with ORCA."""
    
    number = 2
    symmetries = ["Singlet", "Singlet"]
    expected_f_max = 1.0
    expected_l_max = 60000
    # Not sure why this value != 1 for these methods?
    # Perhaps remaining contributions were too small to print? 
    sumofsec = 0.43
    method = "EOM-CCSD"
    
class OrcaADC2Test(OrcaETPostHFTest):
    
    method = "ADC(2)"
    
class OrcaSTEOMCCSDTest(OrcaETPostHFTest):
    """Test for STEOM-CCSD with Orca."""
    
    sumofsec = 1.0
    method = "STEOM-CCSD"
    
class GaussianEOMCCSDTest(GenericTDTest):
    """Test for EOM-CCSD with Gaussian."""
    
    number = 10
    expected_l_max = 61514.3
    expected_f_max = 0.9802
    symmetries = [
            "Triplet-Bu",
            "Triplet-Ag",
            "Triplet-Bu",
            "Singlet-Bu",
            "Triplet-Bu",
            
            "Triplet-Bu",
            "Singlet-Bu",
            "Triplet-Ag",
            "Triplet-Bu",
            "Triplet-Ag",
        ]
    method = "EOM-CCSD"
        

if __name__ =="__main__":

    import sys
    sys.path.insert(1, os.path.join(__filedir__, ".."))

    from test_data import DataSuite
    suite = DataSuite(['TD'])
    suite.testall()
    
